program main

  implicit none

  ! OutASCIIFn  : output file name
  ! Comment     : comment in a file

  character(128) :: OutASCIIFn = 'atmospheric_profile.txt'
  character(128) :: Comment    = 'sample atmospheric profile'


  ! NLev   : Number of levels
  ! NMol   : Number of molecules
  ! NPtcl  : Number of particles

  integer, parameter   :: NLev =  33
  integer, parameter   :: NMol =   3
  integer, parameter   :: NPtcl =  0

  ! Ts                 : surface temperature (K)
  ! r_Press      (:)   : Atmospheric pressure (Pa)
  ! m_MolNum     (:)   : Molecular number
  ! r_Temp       (:)   : Atmospheric temperature (K)
  ! rm_VMR       (:,:) : Volume mixing ratio (1)

  ! a_PtclNum    (:)   : Partile number
  ! a_PtclName   (:)   : Particle name
  ! a_PtclDens   (:)   : Particle density (kg m-3)
  ! za_PtclEffRad(:,:) : Particle effective radius (m)
  ! za_PtclMMR   (:,:) : Particle mass mixing ratio

  real(8)        :: Ts
  real(8)        :: r_Press (NLev)
  integer        :: m_MolNum(NMol)
  real(8)        :: r_Temp  (NLev)
  real(8)        :: rm_VMR  (NLev, NMol)
  !
  integer        :: a_PtclNum    (NPtcl)
  character(128) :: a_PtclName   (NPtcl)
  real(8)        :: a_PtclDens   (NPtcl)
  real(8)        :: za_PtclEffRad(NLev-1, NPtcl)
  real(8)        :: za_PtclMMR   (NLev-1, NPtcl)

  integer :: FileUnit

  integer :: k
  integer :: m



  ! ==========  P(NLev)  , T(NLev)  , VMR(NLev)    (NLev)-th level (upper boundary)
  ! -  -  -  -  Rad(NLev-1), MMR(NLev-1)           (NLev-1)-th layer
  ! ==========  P(NLev-1), T(NLev-1), VMR(NLev-1)  (NLev-1)-th level
  ! -  -  -  -  Rad(NLev-2), MMR(NLev-2)           (NLev-2)-th layer
  !
  ! ...
  !
  ! -  -  -  -  Rad(3), MMR(3)      2nd layer
  ! ==========  P(3), T(3), VMR(3)  3rd level
  ! -  -  -  -  Rad(2), MMR(2)      2nd layer
  ! ==========  P(2), T(2), VMR(2)  2nd level
  ! -  -  -  -  Rad(1), MMR(1)      1st layer
  ! ==========  P(1), T(1), VMR(1)  1st level
  ! ==========  Ts                  ground    (lower boundary)
  !
  !
  ! NLev        : Number of levels (layer interfaces)
  ! NMol        : Number of molecules
  ! MolNumX     : Xth molecular number
  ! Ts          : surface temperature
  ! P(k)        : pressure at k'th level
  ! T(k)        : temperature at k'th level
  ! MolX_VMR(k) : volume mixing ratio of Xth molecule at k'th level
  ! NPtcl       : Number of particles
  ! PtclYName   : name of Yth particle
  ! PtclYDens   : density of Yth particle
  ! PtclYRad(l) : effective radius of of Yth particle at l'th layer
  ! PtclYMMR(l) : mass mixing ratio of of Yth particle at l'th layer


  ! l.1                        : Comment (This line is not read, actually.)
  ! l.2                        : NLev
  ! l.3                        : NMol
  ! l.4                        : MolNum1, MolNum2, ...
  !
  ! l.5                        : Ts
  ! l.6                        : P(1)   , T(1)   , Mol1_VMR(1)   , Mol2_VMR(1)   , ...
  ! l.7                        : P(2)   , T(2)   , Mol1_VMR(2)   , Mol2_VMR(2)   , ...
  ! ...
  ! l.(5+NLev)                 : P(NLev), T(NLev), Mol1_VMR(NLev), Mol2_VMR(NLev), ...
  !
  ! l.(5+NLev+1)               : NPtcl
  ! l.(5+NLev+2)               : Ptcl1Name, Ptcl2Name, ...
  ! l.(5+NLev+3)               : Ptcl1Dens, Ptcl2Dens, ...
  !
  ! l.(5+NLev+3+1)             : Ptcl1Rad(1)     , Ptcl2Rad(1)     , ...
  ! l.(5+NLev+3+2)             : Ptcl1Rad(2)     , Ptcl2Rad(2)     , ...
  ! ...
  ! l.(5+NLev+3+NLev-1)        : Ptcl1Rad(NLev-1), Ptcl2Rad(NLev-1), ...
  !
  ! l.(5+NLev+3+NLev-1+1)      : Ptcl1MMR(1)     , Ptcl2MMR(1) , ...
  ! l.(5+NLev+3+NLev-1+2)      : Ptcl1MMR(2)     , Ptcl2MMR(2) , ...
  ! ...
  ! l.(5+NLev+3+NLev-1+NLev-1) : Ptcl1MMR(NLev-1), Ptcl2MMR(NLev-1) , ...
  !


  ! sample atmospheric profile

  ! Values of r_Press below are those of mid-latitude summer profile.
  data r_Press / &
    & 101300.000   , 90200.0000    , 80200.0000  , 71000.0000  , 62800.0000  ,&
    &  55400.0000  , 48700.0000    , 42600.0000  , 37200.0000  , 32400.0000  ,&
    &  28100.0000  , 24300.0000    , 20900.0000  , 17900.0000  , 15300.0000  ,&
    &  13000.0000  , 11000.0000    ,  9500.00000 ,  8120.00000 ,  6950.00000 ,&
    &   5950.00000 ,  5100.00000   ,  4370.00000 ,  3760.00000 ,  3220.00000 ,&
    &   2770.00000 ,  1320.00000   ,   652.000000,   333.000000,   176.000000,&
    &     95.0999985,    6.71000004, 2.99999993E-02 &
    & /

  !   Pressure level is equally spaced in logarithmic scale with an interval
  !   of 1/10 scale height.
!!$  do k = 1, NLev
!!$    r_Press(k) = 1.0d5 * exp( - 1.0d0/10.0d0 * ( k - 1 ) )
!!$  end do

  Ts = 294.0d0

  ! Values of rm_VMR below are those of mid-latitude summer profile.
  data r_Temp / &
    & 294.000000, 290.000000, 285.000000, 279.000000, 273.000000, &
    & 267.100006, 261.000000, 254.699997, 248.199997, 241.699997, &
    & 235.199997, 228.800003, 222.300003, 216.899994, 215.800003, &
    & 215.800003, 215.800003, 215.800003, 216.000000, 217.000000, &
    & 218.199997, 219.399994, 220.600006, 221.800003, 223.000000, &
    & 224.199997, 234.199997, 245.300003, 257.500000, 269.700012, &
    & 276.200012, 219.100006, 209.899994 &
    & /
  !   Isothermal atmosphere of 294 K.
!!$  do k = 1, NLev
!!$    r_Temp(k) = 294.0d0
!!$  end do


  m_MolNum(1) = 1  ! H2O
  m_MolNum(2) = 2  ! CO2
  m_MolNum(3) = 3  ! O3

  ! Values of rm_VMR below are those of mid-latitude summer profile.
  data rm_VMR / &
    ! H2O
    & 1.87514704E-02, 1.37988785E-02, 9.59391333E-03, 6.22027181E-03, &
    & 3.79170198E-03, 2.22501787E-03, 1.50625419E-03, 1.02367636E-03, &
    & 6.46617671E-04, 4.06240375E-04, 2.48377008E-04, 9.51619368E-05, &
    & 3.17099148E-05, 9.28290683E-06, 6.47668048E-06, 6.43512158E-06, &
    & 6.42815712E-06, 6.43673457E-06, 6.43277599E-06, 6.42656914E-06, &
    & 6.43118165E-06, 6.43251587E-06, 6.42987561E-06, 6.42473333E-06, &
    & 6.42413215E-06, 6.42471196E-06, 6.42764689E-06, 6.42423038E-06, &
    & 6.42354735E-06, 6.42838040E-06, 6.43359635E-06, 6.43453450E-06, &
    & 6.42559962E-06, &
    ! CO2
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, 3.00000014E-04, &
    & 3.00000014E-04, &
    ! O3
    & 3.01636724E-08, 3.34147146E-08, 3.69331907E-08, 4.22019610E-08, &
    & 4.81923443E-08, 5.51192265E-08, 6.40553921E-08, 7.76740023E-08, &
    & 9.13021552E-08, 1.11128429E-07, 1.30487408E-07, 1.79406229E-07, &
    & 2.21090204E-07, 3.14842168E-07, 4.39772037E-07, 5.46332160E-07, &
    & 7.13630072E-07, 9.44352564E-07, 1.29018156E-06, 1.73069282E-06, &
    & 2.15979117E-06, 2.68265012E-06, 3.14790509E-06, 3.47414289E-06, &
    & 3.83878705E-06, 4.20602828E-06, 6.14664350E-06, 5.99560281E-06, &
    & 5.49176502E-06, 3.45069611E-06, 2.16324997E-06, 4.86422209E-07, &
    & 5.21140358E-08  &
    & /

  !   Following molecules are included:
  !     H2O of 1.0 ppmv,
  !     CO2 of 300 ppmv,
  !     O3  of 0.0 ppmv.
!!$  do k = 1, NLev
!!$    ! H2O
!!$    rm_VMR(k,1) =   1.0d-6
!!$    ! CO2
!!$    rm_VMR(k,2) = 300.0d-6
!!$    ! O3
!!$    rm_VMR(k,3) =   0.0d0
!!$  end do



  open( FileUnit, file = OutASCIIFN, status = 'unknown' )
  write( FileUnit, * ) trim( Comment )
  write( FileUnit, * ) NLev
  write( FileUnit, * ) NMol
  write( FileUnit, * ) ( m_MolNum(m), m=1,NMol )
  write( FileUnit, * ) Ts
  do k = 1, NLev
    write( FileUnit, * ) r_Press(k), r_Temp(k), ( rm_VMR(k,m), m=1,NMol )
  end do
  write( FileUnit, * ) NPtcl
  if ( NPtcl > 0 ) then
    do m = 1, NPtcl
      a_PtclNum(m) = m
    end do
    write( FileUnit, * ) ( a_PtclName  (m), m=1,NPtcl )
    write( FileUnit, * ) ( a_PtclDens  (m), m=1,NPtcl )
    do k = 1, NLev-1
      write( FileUnit, * ) ( za_PtclEffRad(k,m), m=1,NPtcl )
    end do
    do k = 1, NLev-1
      write( FileUnit, * ) ( za_PtclMMR(k,m), m=1,NPtcl )
    end do
  end if
  close( FileUnit )


  write( 6, * ) 'Finish'

end program main

!------------------------------------------------------------------------------
