program mk2Dnc

  use vtype_module
  use ni3_module

!!$  integer, parameter :: NLat   = 6
  integer, parameter :: NLat   = 19
  integer, parameter :: NPress = 80

  real(dp) :: y_Lat( NLat )

  real(dp) :: z_Press( NPress )
  real(dp) :: z_Temp ( NPress )
  real(dp) :: Temp
  real(dp) :: yz_Temp( NLat, NPress )

  real(dp), parameter :: FillValue = -1.0d10

  real(dp) :: z_PressTmp( NPress )

  character(extstr) :: NcFNIn
  character(extstr) :: NcFNOut = "out/Venus_VIRA_2D.nc"
  integer           :: NcIDIn
  integer           :: NcIDOut

  character(stdstr) :: Mode
  character(stdstr) :: Name
  character(stdstr) :: StdName
  character(stdstr) :: LongName
  character(stdstr) :: Unit
  integer           :: NDims
  character(stdstr), allocatable :: a_DimNames( : )

  integer :: j
  integer :: k
  integer :: kk


!!$  data y_Lat / 0, 30, 45, 60, 75, 85 /
  data y_Lat / 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 89 /


  ! output file
  Mode = "new"
  call ni3_open( NcFNOut, Mode, NcIDOut )

  do j = 1, NLat

    write( NcFNIn, '(a,i2.2,a)' ) &
      & '../out/Venus_VIRA_', int(y_Lat(j)), '+GasVMR+CldMMR+UVabsMMR_H2O_CO2_CO_SO2_HF_OCS_N2.nc'

    Mode = "read"
    call ni3_open( NcFNIn, Mode, NcIDIn )
    call ni3_get_var( NcIDIn, "Press", z_PressTmp )
    if ( j == 1 ) then
      z_Press = z_PressTmp
    end if

    call ni3_get_var( NcIDIn, "Temp", z_Temp )
    do k = 1, NPress

      if ( z_Press(k) > z_PressTmp(1) ) then
        Temp = FillValue
      else if ( z_Press(k) < z_PressTmp(NPress) ) then
        Temp = FillValue
      else
        do kk = 1+1, NPress
          if ( z_PressTmp(kk) <= z_Press(k) ) then
            exit
          end if
        end do
        Temp =   ( z_Temp(kk) - z_Temp(kk-1) ) &
          &  / log( z_PressTmp(kk) / z_PressTmp(kk-1) ) &
          &  * log( z_Press(k)     / z_PressTmp(kk-1) ) &
          &  + z_Temp(kk-1)
      end if

      yz_Temp(j,k) = Temp
    end do

    ! preparation of the output file
    if ( j == 1 ) then
      Name = "lat"
      StdName  = Name
      LongName = "latitude"
      Unit     = "degrees_north"
      call ni3_set_dim( NcIDOut, Name, NI3_DOUBLE, y_Lat, &
        & StdName, LongName, Unit )
      !
      Name = "Press"
      call ni3_set_dim( NcIDOut, Name, NI3_DOUBLE, z_Press )
      call ni3_cp_atts( NcIDIn, NcIDOut, Name )
      !
      NDims = 2
      allocate( a_DimNames( NDims ) )
      a_DimNames(1) = 'lat'
      a_DimNames(2) = 'Press'
      Name = "Temp"
      call ni3_def_var( NcIDOut, Name, NI3_DOUBLE, NDims, a_DimNames, &
        & fillvalue = FillValue )
      call ni3_put_att( NcIDOut, Name, "missing_value", FillValue )
      deallocate( a_DimNames )
    end if


    call ni3_close( NcIDIn )

  end do

  call ni3_put_var( NcIDOut, "Temp", yz_Temp )
  call ni3_close( NcIDOut )


end program mk2Dnc
