module main_output

  use vtype_module

  implicit none

  private

  public :: CalcAndOutputFlux
  public :: EstimateError


  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------

  subroutine CalcAndOutputFlux( &
    & Grav, Cp, &
    & Method, InNcFN, InPtclOptPropNcFN, InStrSpeNcFN, OutNcFN1, OutNcFN2, &
    & kmax, NMol, m_MolNum, NPtcl, a_PtclName, &
    & r_Press, z_Press, &
    & SurfTemp, r_Temp, rm_VMR, r_MeanMolWt, &
    & za_PtclDens, za_PtclEffRad, za_DelPtclMass, &
    & WaveNumAlbedoSwitch, AlbedoSW, AlbedoLW, EmisSW, EmisLW, &
    & NGQ, a_GQWeight, NGQLoop, s_SZADeg, &
    & StrFluxTOA, &
    & WNIntegStart, WNIntegEnd, &
    & r_RadUwFluxLW, r_RadDwFluxLW, z_DTempDtLW, &
    & r_RadUwFluxSW, r_RadDwFluxSW, z_DTempDtSW, &
    & r_RadUwFlux  , r_RadDwFlux  , z_DTempDt    &
    & )

    use rad_flux, only : &
      & RadFluxCalcFluxMultiSZA, RadFluxCalcFlux, &
      & RadFluxInit, RadFluxFinalize, RadFluxOutputInit, RadFluxOutputFinalize

    use rad_kd_utils, only : RadKDUtilsCalcTendency

    use rad_output, only : RadOutputFluxTendency

    real(DP)    , intent(in ) :: Grav
    real(DP)    , intent(in ) :: Cp
    character(*), intent(in ) :: Method
    character(*), intent(in ) :: InNcFN
    character(*), intent(in ) :: InPtclOptPropNcFN
    character(*), intent(in ) :: InStrSpeNcFN
    character(*), intent(in ) :: OutNcFN1
    character(*), intent(in ) :: OutNcFN2
    integer     , intent(in ) :: kmax
    integer     , intent(in ) :: NMol
    integer     , intent(in ) :: m_MolNum   (1:NMol)
    integer     , intent(in ) :: NPtcl
    character(*), intent(in ) :: a_PtclName(NPtcl)
    real(DP)    , intent(in ) :: r_Press    (0:kmax)
    real(DP)    , intent(in ) :: z_Press    (1:kmax)
    real(DP)    , intent(in ) :: SurfTemp
    real(DP)    , intent(in ) :: r_Temp     (0:kmax)
    real(DP)    , intent(in ) :: rm_VMR     (0:kmax,NMol)
    real(DP)    , intent(in ) :: r_MeanMolWt(0:kmax)
    real(DP)    , intent(in ) :: za_PtclDens   (kmax,NPtcl)
    real(DP)    , intent(in ) :: za_PtclEffRad (kmax,NPtcl)
    real(DP)    , intent(in ) :: za_DelPtclMass(kmax,NPtcl)
    real(DP)    , intent(in ) :: WaveNumAlbedoSwitch
    real(DP)    , intent(in ) :: AlbedoSW
    real(DP)    , intent(in ) :: AlbedoLW
    real(DP)    , intent(in ) :: EmisSW
    real(DP)    , intent(in ) :: EmisLW
    integer     , intent(in ) :: NGQ
    real(DP)    , intent(in ) :: a_GQWeight( NGQ+1 )
    integer     , intent(in ) :: NGQLoop
    real(DP)    , intent(in ) :: s_SZADeg(1:NGQLoop)
    real(DP)    , intent(in ) :: StrFluxTOA
    real(DP)    , intent(in ) :: WNIntegStart
    real(DP)    , intent(in ) :: WNIntegEnd
    real(DP)    , intent(out) :: r_RadUwFluxLW    (0:kmax)
    real(DP)    , intent(out) :: r_RadDwFluxLW    (0:kmax)
    real(DP)    , intent(out) :: z_DTempDtLW      (1:kmax)
    real(DP)    , intent(out) :: r_RadUwFluxSW    (0:kmax)
    real(DP)    , intent(out) :: r_RadDwFluxSW    (0:kmax)
    real(DP)    , intent(out) :: z_DTempDtSW      (1:kmax)
    real(DP)    , intent(out) :: r_RadUwFlux      (0:kmax)
    real(DP)    , intent(out) :: r_RadDwFlux      (0:kmax)
    real(DP)    , intent(out) :: z_DTempDt        (1:kmax)


    ! Local variables
    !
    integer , parameter   :: NDelFlxMax = 1

    real(DP) :: rs_RadUwFluxEach      (0:kmax,1:NGQLoop)
    real(DP) :: rs_RadDwFluxEach      (0:kmax,1:NGQLoop)
    real(DP) :: ras_DelRadUwFluxEach  (0:kmax,-1:NDelFlxMax,1:NGQLoop)
    real(DP) :: ras_DelRadDwFluxEach  (0:kmax,-1:NDelFlxMax,1:NGQLoop)
    real(DP) :: rs_RadUwFluxSWEach    (0:kmax,1:NGQLoop)
    real(DP) :: rs_RadDwFluxSWEach    (0:kmax,1:NGQLoop)
    real(DP) :: ras_DelRadUwFluxSWEach(0:kmax,-1:NDelFlxMax,1:NGQLoop)
    real(DP) :: ras_DelRadDwFluxSWEach(0:kmax,-1:NDelFlxMax,1:NGQLoop)

    real(DP) :: ra_DelRadUwFluxLW(0:kmax,-1:NDelFlxMax)
    real(DP) :: ra_DelRadDwFluxLW(0:kmax,-1:NDelFlxMax)
    real(DP) :: ra_DelRadUwFluxSW(0:kmax,-1:NDelFlxMax)
    real(DP) :: ra_DelRadDwFluxSW(0:kmax,-1:NDelFlxMax)
    real(DP) :: ra_DelRadUwFlux  (0:kmax,-1:NDelFlxMax)
    real(DP) :: ra_DelRadDwFlux  (0:kmax,-1:NDelFlxMax)

    integer  :: n


    call RadFluxInit( &
      & Grav, &
      & WaveNumAlbedoSwitch, &
      & InNcFN, InPtclOptPropNcFN, &
      & InStrSpeNcFN, &
      & Method  &
      & )
    call RadFluxCalcFluxMultiSZA( &
      & kmax, NMol, &
      & m_MolNum, &
      & r_Press, &
      & SurfTemp, r_Temp, rm_VMR, &
      & NPtcl, a_PtclName, za_PtclDens, za_PtclEffRad, za_DelPtclMass, &
      & AlbedoSW, AlbedoLW, &
      & EmisSW, EmisLW, &
      & NGQLoop, s_SZADeg, &
      & StrFluxTOA, &
      & r_MeanMolWt, &
      & NDelFlxMax, &
      & rs_RadUwFluxEach      , rs_RadDwFluxEach      , &
      & ras_DelRadUwFluxEach  , ras_DelRadDwFluxEach  , &
      & r_RadUwFluxLW         , r_RadDwFluxLW         , &
      & rs_RadUwFluxSWEach    , rs_RadDwFluxSWEach    , &
      & ra_DelRadUwFluxLW     , ra_DelRadDwFluxLW     , &
      & ras_DelRadUwFluxSWEach, ras_DelRadDwFluxSWEach, &
      & WNIntegStart, WNIntegEnd &
      & )
    r_RadUwFlux       = 0.0d0
    r_RadDwFlux       = 0.0d0
    ra_DelRadUwFlux   = 0.0d0
    ra_DelRadDwFlux   = 0.0d0
    r_RadUwFluxSW     = 0.0d0
    r_RadDwFluxSW     = 0.0d0
    ra_DelRadUwFluxSW = 0.0d0
    ra_DelRadDwFluxSW = 0.0d0
    do n = 1, NGQLoop
      r_RadUwFlux       = r_RadUwFlux       + rs_RadUwFluxEach      (:,n)   * a_GQWeight(n)
      r_RadDwFlux       = r_RadDwFlux       + rs_RadDwFluxEach      (:,n)   * a_GQWeight(n)
      ra_DelRadUwFlux   = ra_DelRadUwFlux   + ras_DelRadUwFluxEach  (:,:,n) * a_GQWeight(n)
      ra_DelRadDwFlux   = ra_DelRadDwFlux   + ras_DelRadDwFluxEach  (:,:,n) * a_GQWeight(n)
      r_RadUwFluxSW     = r_RadUwFluxSW     + rs_RadUwFluxSWEach    (:,n)   * a_GQWeight(n)
      r_RadDwFluxSW     = r_RadDwFluxSW     + rs_RadDwFluxSWEach    (:,n)   * a_GQWeight(n)
      ra_DelRadUwFluxSW = ra_DelRadUwFluxSW + ras_DelRadUwFluxSWEach(:,:,n) * a_GQWeight(n)
      ra_DelRadDwFluxSW = ra_DelRadDwFluxSW + ras_DelRadDwFluxSWEach(:,:,n) * a_GQWeight(n)
    end do
    call RadFluxFinalize

    call RadKDUtilsCalcTendency( &
      & Grav, Cp, &
      & kmax, r_Press, &
      & r_RadUwFlux, r_RadDwFlux, &
      & z_DTempDt &
      & )
    call RadKDUtilsCalcTendency( &
      & Grav, Cp, &
      & kmax, r_Press, &
      & r_RadUwFluxLW, r_RadDwFluxLW, &
      & z_DTempDtLW &
      & )
    call RadKDUtilsCalcTendency( &
      & Grav, Cp, &
      & kmax, r_Press, &
      & r_RadUwFluxSW, r_RadDwFluxSW, &
      & z_DTempDtSW &
      & )
    call RadOutputFluxTendency( &
      & OutNcFN1, OutNcFN2,   &
      & kmax, r_Press, z_Press, &
      & r_RadUwFlux, r_RadDwFlux, z_DTempDt, &
      & r_RadUwFluxLW, r_RadDwFluxLW, z_DTempDtLW, &
      & r_RadUwFluxSW, r_RadDwFluxSW, z_DTempDtSW, &
      & WNStart = WNIntegStart, WNEnd = WNIntegEnd &
      & )


  end subroutine CalcAndOutputFlux

  !----------------------------------------------------------------------------

  subroutine EstimateError(&
    & FileUnitQL, &
    & NBandCur, ab_BandCurWNBnds, &
    & iBand, iMode, &
    & kmax, &
    & r_RadUwFlux     , r_RadDwFlux     , &
    & r_RadUwFluxLW   , r_RadDwFluxLW   , &
    & r_RadUwFluxSW   , r_RadDwFluxSW   , &
    & r_RadUwFluxRef  , r_RadDwFluxRef  , &
    & r_RadUwFluxLWRef, r_RadDwFluxLWRef, &
    & r_RadUwFluxSWRef, r_RadDwFluxSWRef, &
    & r_RadUwFluxDenom  , r_RadDwFluxDenom,   &
    & r_RadUwFluxLWDenom, r_RadDwFluxLWDenom, &
    & r_RadUwFluxSWDenom, r_RadDwFluxSWDenom  &
    & )

    integer , intent(in ) :: FileUnitQL
    integer , intent(in ) :: NBandCur
    real(DP), intent(in ) :: ab_BandCurWNBnds (2,NBandCur)
    integer , intent(in ) :: iBand
    integer , intent(in ) :: iMode
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_RadUwFlux     (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux     (0:kmax)
    real(dp), intent(in ) :: r_RadUwFluxLW   (0:kmax)
    real(dp), intent(in ) :: r_RadDwFluxLW   (0:kmax)
    real(dp), intent(in ) :: r_RadUwFluxSW   (0:kmax)
    real(dp), intent(in ) :: r_RadDwFluxSW   (0:kmax)
    real(dp), intent(in ) :: r_RadUwFluxRef  (0:kmax)
    real(dp), intent(in ) :: r_RadDwFluxRef  (0:kmax)
    real(dp), intent(in ) :: r_RadUwFluxLWRef(0:kmax)
    real(dp), intent(in ) :: r_RadDwFluxLWRef(0:kmax)
    real(dp), intent(in ) :: r_RadUwFluxSWRef(0:kmax)
    real(dp), intent(in ) :: r_RadDwFluxSWRef(0:kmax)
    real(dp), intent(in ), optional :: r_RadUwFluxDenom  (0:kmax)
    real(dp), intent(in ), optional :: r_RadDwFluxDenom  (0:kmax)
    real(dp), intent(in ), optional :: r_RadUwFluxLWDenom(0:kmax)
    real(dp), intent(in ), optional :: r_RadDwFluxLWDenom(0:kmax)
    real(dp), intent(in ), optional :: r_RadUwFluxSWDenom(0:kmax)
    real(dp), intent(in ), optional :: r_RadDwFluxSWDenom(0:kmax)


    ! Local variables
    !
    real(dp) :: RadUwFluxErrMax
    real(dp) :: RadDwFluxErrMax
    real(dp) :: RadNetFluxErrMax
    real(dp) :: RadUwFluxLWErrMax
    real(dp) :: RadDwFluxLWErrMax
    real(dp) :: RadNetFluxLWErrMax
    real(dp) :: RadUwFluxSWErrMax
    real(dp) :: RadDwFluxSWErrMax
    real(dp) :: RadNetFluxSWErrMax

!!$    real(dp) :: RadUwFluxRatioMax
!!$    real(dp) :: RadDwFluxRatioMax
!!$    real(dp) :: RadUwFluxLWRatioMax
!!$    real(dp) :: RadDwFluxLWRatioMax
!!$    real(dp) :: RadUwFluxSWRatioMax
!!$    real(dp) :: RadDwFluxSWRatioMax

    integer :: i
    integer :: FU

    real(dp) :: FluxThreshold = 1.0d-1


    ! relative error
    RadUwFluxErrMax   = RelErrMax( kmax, r_RadUwFlux  , r_RadUwFluxRef  , r_RadUwFluxDenom  , FluxThreshold )
    RadDwFluxErrMax   = RelErrMax( kmax, r_RadDwFlux  , r_RadDwFluxRef  , r_RadDwFluxDenom  , FluxThreshold )
    RadUwFluxLWErrMax = RelErrMax( kmax, r_RadUwFluxLW, r_RadUwFluxLWRef, r_RadUwFluxLWDenom, FluxThreshold )
    RadDwFluxLWErrMax = RelErrMax( kmax, r_RadDwFluxLW, r_RadDwFluxLWRef, r_RadDwFluxLWDenom, FluxThreshold )
    RadUwFluxSWErrMax = RelErrMax( kmax, r_RadUwFluxSW, r_RadUwFluxSWRef, r_RadUwFluxSWDenom, FluxThreshold )
    RadDwFluxSWErrMax = RelErrMax( kmax, r_RadDwFluxSW, r_RadDwFluxSWRef, r_RadDwFluxSWDenom, FluxThreshold )

    if ( present( r_RadUwFluxDenom   ) .and. present( r_RadUwFluxDenom   ) .and. &
      &  present( r_RadUwFluxLWDenom ) .and. present( r_RadUwFluxLWDenom ) .and. &
      &  present( r_RadUwFluxSWDenom ) .and. present( r_RadUwFluxSWDenom ) ) then
      RadNetFluxErrMax   = &
        & RelErrMax( kmax, r_RadUwFlux  -r_RadDwFlux  , r_RadUwFluxRef  -r_RadDwFluxRef  , &
        &            r_RadUwFluxDenom  -r_RadDwFluxDenom, FluxThreshold )
      RadNetFluxLWErrMax = &
        & RelErrMax( kmax, r_RadUwFluxLW-r_RadDwFluxLW, r_RadUwFluxLWRef-r_RadDwFluxLWRef, &
        &            r_RadUwFluxLWDenom-r_RadDwFluxLWDenom, FluxThreshold )
      RadNetFluxSWErrMax = &
        & RelErrMax( kmax, r_RadUwFluxSW-r_RadDwFluxSW, r_RadUwFluxSWRef-r_RadDwFluxSWRef, &
        &            r_RadUwFluxSWDenom-r_RadDwFluxSWDenom, FluxThreshold )
    else
      RadNetFluxErrMax   = &
        & RelErrMax( kmax, r_RadUwFlux  -r_RadDwFlux  , r_RadUwFluxRef  -r_RadDwFluxRef  , &
        &            FluxThreshold=FluxThreshold )
      RadNetFluxLWErrMax = &
        & RelErrMax( kmax, r_RadUwFluxLW-r_RadDwFluxLW, r_RadUwFluxLWRef-r_RadDwFluxLWRef, &
        &            FluxThreshold=FluxThreshold )
      RadNetFluxSWErrMax = &
        & RelErrMax( kmax, r_RadUwFluxSW-r_RadDwFluxSW, r_RadUwFluxSWRef-r_RadDwFluxSWRef, &
        &            FluxThreshold=FluxThreshold )
    end if


    ! ratio
!!$    RadUwFluxRatioMax   = RatioMax( kmax, r_RadUwFlux  , r_RadUwFluxRef  , FluxThreshold )
!!$    RadDwFluxRatioMax   = RatioMax( kmax, r_RadDwFlux  , r_RadDwFluxRef  , FluxThreshold )
!!$    RadUwFluxLWRatioMax = RatioMax( kmax, r_RadUwFluxLW, r_RadUwFluxLWRef, FluxThreshold )
!!$    RadDwFluxLWRatioMax = RatioMax( kmax, r_RadDwFluxLW, r_RadDwFluxLWRef, FluxThreshold )
!!$    RadUwFluxSWRatioMax = RatioMax( kmax, r_RadUwFluxSW, r_RadUwFluxSWRef, FluxThreshold )
!!$    RadDwFluxSWRatioMax = RatioMax( kmax, r_RadDwFluxSW, r_RadDwFluxSWRef, FluxThreshold )


    do i = 1, 2
      select case ( i )
      case ( 1 )
        FU = 6
      case ( 2 )
        FU = FileUnitQL
      end select
      write( FU, * ) 'iBand = ', iBand, 'iMode = ', iMode
      write( FU, * ) '   ', 'Wavenumber = ', int(ab_BandCurWnBnds(1,1)), '-', int(ab_BandCurWnBnds(2,NBandCur)), '(m-1)'
      write( FU, * ) '   ', 'Total'
      write( FU, * ) '   ', '  Net flux : ', 'max rel. err. = ', RadNetFluxErrMax
      write( FU, * ) '   ', '  Up. flux : ', 'max rel. err. = ', RadUwFluxErrMax!, ', max rat. = ', RadUwFluxRatioMax
      write( FU, * ) '   ', '  Dn. flux : ', 'max rel. err. = ', RadDwFluxErrMax!, ', max rat. = ', RadDwFluxRatioMax
      write( FU, * ) '   ', 'LW'
      write( FU, * ) '   ', '  Net flux : ', 'max rel. err. = ', RadNetFluxLWErrMax
      write( FU, * ) '   ', '  Up. flux : ', 'max rel. err. = ', RadUwFluxLWErrMax!, ', max rat. = ', RadUwFluxLWRatioMax
      write( FU, * ) '   ', '  Dn. flux : ', 'max rel. err. = ', RadDwFluxLWErrMax!, ', max rat. = ', RadDwFluxLWRatioMax
      write( FU, * ) '   ', 'SW'
      write( FU, * ) '   ', '  Net flux : ', 'max rel. err. = ', RadNetFluxSWErrMax
      write( FU, * ) '   ', '  Up. flux : ', 'max rel. err. = ', RadUwFluxSWErrMax!, ', max rat. = ', RadUwFluxSWRatioMax
      write( FU, * ) '   ', '  Dn. flux : ', 'max rel. err. = ', RadDwFluxSWErrMax!, ', max rat. = ', RadDwFluxSWRatioMax
      call flush( FU )
    end do


  end subroutine EstimateError

  !----------------------------------------------------------------------------

  function RelErrMax( kmax, r_RadFlux, r_RadFluxRef, r_RadFluxDenom, FluxThreshold ) &
    result( RadFluxErrMax )

    integer , intent(in )           :: kmax
    real(dp), intent(in )           :: r_RadFlux     (0:kmax)
    real(dp), intent(in )           :: r_RadFluxRef  (0:kmax)
    real(dp), intent(in ), optional :: r_RadFluxDenom(0:kmax)
    real(dp), intent(in ), optional :: FluxThreshold

    real(dp) :: RadFluxErrMax


    ! Local variables
    !
    real(dp) :: r_RadFluxDenomLV(0:kmax)
    real(dp) :: FluxThresholdLV
    real(dp) :: RelErr
    integer :: k


    if ( present( r_RadFluxDenom ) ) then
      r_RadFluxDenomLV = r_RadFluxDenom + 1.0d-100
    else
      r_RadFluxDenomLV = r_RadFluxRef   + 1.0d-100
    end if
    if ( present( FluxThreshold ) ) then
      FluxThresholdLV = FluxThreshold
    else
      FluxThresholdLV = -1.0d0
    end if

    RadFluxErrMax   = 0.0d0
    do k = kmax, 0, -1
      if ( abs(r_RadFluxRef(k)) > FluxThresholdLV ) then
!!$        RadFluxErrMax = &
!!$          & max( RadFluxErrMax, &
!!$          &   abs( ( r_RadFlux(k) - r_RadFluxRef(k) ) / r_RadFluxDenomLV(k) ) )
        RelErr = ( r_RadFlux(k) - r_RadFluxRef(k) ) / r_RadFluxDenomLV(k)
        if ( abs( RelErr ) > abs( RadFluxErrMax ) ) then
          RadFluxErrMax = RelErr
        end if
      end if
    end do


  end function RelErrMax

  !----------------------------------------------------------------------------

  function RatioMax( kmax, r_RadFlux, r_RadFluxRef, FluxThreshold ) &
    result( RadFluxRatioMax )

    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_RadFlux   (0:kmax)
    real(dp), intent(in ) :: r_RadFluxRef(0:kmax)
    real(dp), intent(in ) :: FluxThreshold

    real(dp) :: RadFluxRatioMax


    ! Local variables
    !
    integer :: k


    RadFluxRatioMax = 1.0d0
    do k = kmax, 0, -1
      if ( abs( r_RadFluxRef(k) ) > FluxThreshold ) then
        RadFluxRatioMax = &
          & max( &
          &      abs(RadFluxRatioMax-1.0d0), &
          &      abs( abs( r_RadFlux(k) / (r_RadFluxRef(k)+1.0d-100) ) - 1.0d0 ) &
          &    ) &
          & + 1.0d0
      end if
    end do


  end function RatioMax

  !----------------------------------------------------------------------------

end module main_output
