program main

  use vtype_module

  use fi_module

  use io_profile, only : &
    & IOProfileInqDimLen, &
    & IOProfileRead, &
    & IOProfileReadFlag

  use ac_calc, only : ACCalc

  use ac_calc_utils, only : &
    & ACCalcUtilsModifyWavenumberRange1, &
    & ACCalcUtilsModifyWavenumberRange2, &
    & ACCalcUtilsModifyWavenumberRange3

#ifdef MPI
  use mpi
#endif

  implicit none


  character(extstr) :: CtlFN  = 'calc_ac.conf'
  character(extstr) :: CtlFN2 = 'file.conf'
  character(extstr) :: CtlFN3 = 'many.conf'
  character(extstr) :: CtlFN4 = 'mpisubloop.conf'

  integer           :: iostat

  logical, allocatable :: m_FlagCalcLine(:)
  logical, allocatable :: m_FlagCalcCont(:)

  integer, parameter :: MaxNmlArySize = 64


  logical                            :: FlagNetCDF

  character(len=extstr)              :: HITFileName(MaxNmlArySize)
  integer                            :: NHIT
  character(len=extstr), allocatable :: a_HITFileName(:)
  character(len=extstr)              :: OutDir
  character(len=extstr)              :: OutputFileName
  character(len=extstr)              :: OutputFileNameTmp


  integer                       :: NLev

  !
  ! NMolNum : number of radiatively active species
  !
  integer                        :: NMolNum


  real(dp)                           :: DelWaveNum
  real(dp)                           :: WaveNumS
  real(dp)                           :: WaveNumE

  real(dp)                           :: WaveNumSSave
  real(dp)                           :: WaveNumESave


  integer(i4b)              :: FileUnit



  character(len=extstr)     :: FileNameTail

  character(len=extstr)     :: InFileName
  character(len=extstr)     :: PressName
  character(len=extstr)     :: MolNumName
  character(len=extstr)     :: TempName
  character(len=extstr)     :: SurfTempName
  character(len=extstr)     :: VMRName

  real(dp)    , allocatable :: z_Press(:)
  integer     , allocatable :: m_MolNum( : )

  real(dp)    , allocatable :: z_Temp(:)
  real(dp)                  :: SurfTemp

  logical     , allocatable :: z_FlagCalc(:)


  !
  ! zm_VMR       ( NLev, NMolNum ) :: Volume Mixing Ratio
  !
  real(dp)    , allocatable :: zm_VMR(:,:)
  real(dp)    , allocatable :: zm_VMRShrink(:,:)

  logical  :: FlagCalcLine(MaxNmlArySize)
  logical  :: FlagCalcCont(MaxNmlArySize)

  real(dp)              :: LineCutOffWaveNum(MaxNmlArySize)
  real(dp), allocatable :: m_LineCutOffWaveNum(:)

  real(dp), allocatable :: m_AirBroadHalfWidFac(:)
  real(dp)              :: AirBroadHalfWidFac(MaxNmlArySize)

  character(len=extstr)              :: CIAFileName(MaxNmlArySize)
  integer                            :: NCIA
  character(len=extstr), allocatable :: a_CIAFileName(:)

  character(len=extstr)              :: XSCFileName(MaxNmlArySize)
  integer                            :: NXSC
  character(len=extstr), allocatable :: a_XSCFileName(:)

  integer :: iTemp
  integer :: iTemps
  integer :: iTempe
  integer :: Mol1Num
  integer :: iVMRMol1
  integer :: iVMRMol1s
  integer :: iVMRMol1e
  integer :: Mol2Num
  integer :: iVMRMol2
  integer :: iVMRMol2s
  integer :: iVMRMol2e
  integer :: Mol3Num
  integer :: iVMRMol3
  integer :: iVMRMol3s
  integer :: iVMRMol3e

  integer :: NTemp
  integer :: NVMRMol1
  integer :: NVMRMol2
  integer :: NVMRMol3

  integer :: iLoop
  integer :: NLoop
  integer :: iLoops
  integer :: iLoope

  integer :: iRank

  integer :: IErr
  integer :: NProcs
  integer :: MyRank

  integer            :: NProcsMPISubLoop
  integer            :: iProcWaveNumLoop
  integer, parameter :: FileNumS = 0

  integer :: ks
  integer :: ke

  integer :: k
  integer :: m
  integer :: n


  namelist /wn_nml/ &
    & DelWaveNum, WaveNumS, WaveNumE
  namelist /gas_nml/ &
    & FlagCalcLine, FlagCalcCont, &
    & FlagNetCDF, HITFileName, CIAFileName, XSCFileName, &
    & LineCutOffWaveNum, &
    & AirBroadHalfWidFac
  namelist /file_nml/ &
    & OutDir
!!$    & InFileName, &
!!$    & OutputFileName, &
!!$    & PressName, MolNumName, TempName, VMRName
  namelist /many_nml/ &
    & iTemps, iTempe, &
    & Mol1Num, iVMRMol1s, iVMRMol1e, &
    & Mol2Num, iVMRMol2s, iVMRMol2e, &
    & Mol3Num, iVMRMol3s, iVMRMol3e
  namelist /mpisubloop_nml/ &
    & NProcsMPISubLoop



#ifdef MPI
  call mpi_init( IErr )
  call mpi_comm_size( mpi_comm_world, NProcs, IErr )
  call mpi_comm_rank( mpi_comm_world, MyRank, IErr )
#else
  NProcs = 1
  MyRank = 0
#endif


  FlagCalcLine(:) = .true.
  FlagCalcCont(:) = .true.

  FlagNetCDF = .true.
  HITFileName(:) = ""
  !
  CIAFileName(:) = ""
!  CIAFileName(1) = "../data_hitran/CIA/NC_data/02-02_cia.nc"
  !
  XSCFileName(:) = ""
  !
  OutDir         = "out"

!!$      ! for 15 mincron band
!!$      WaveNumS =  400.0d0 * 1.0d2
!!$      WaveNumE =  900.0d0 * 1.0d2
!!$      ! for CO2 500-900 cm^-1 for validation
!!$      WaveNumS =  500.0d0 * 1.0d2
!!$      WaveNumE =  900.0d0 * 1.0d2
  ! for 15 mincron band for validation
  DelWaveNum = 1.0d-2 * 1.0d2
  WaveNumS =  550.0d0 * 1.0d2 + DelWaveNum / 2.0d0
  WaveNumE =  750.0d0 * 1.0d2 - DelWaveNum / 2.0d0

  WaveNumS =  600.0d0 * 1.0d2 + DelWaveNum / 2.0d0
  WaveNumE =  601.0d0 * 1.0d2 - DelWaveNum / 2.0d0


!!$      WaveNumS =    0.0d0 * 1.0d2
!!$      WaveNumE = 3000.0d0 * 1.0d2

  LineCutOffWaveNum(:) = 25.0d2

  AirBroadHalfWidFac = 1.0d0

  !======================
  InFileName   = "./data.nc.nc"
  PressName    = "Press"
  MolNumName   = "MolNum"
  TempName     = "Temp"
  SurfTempName = "Ts"
  VMRName      = "VMR"
  !======================


  ! Read a namelist file

  call fi_open( CtlFN, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = wn_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in wn_nml is something wrong.'
  end if
  write( 6, nml = wn_nml )
  !
  rewind( FileUnit )
  read( FileUnit, nml = gas_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in gas_nml is something wrong.'
  end if
  write( 6, nml = gas_nml )
  !
  close( FileUnit )

  call fi_open( CtlFN2, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = file_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in file_nml is something wrong.'
  end if
  write( 6, nml = file_nml )
  !
  close( FileUnit )

  call fi_open( CtlFN3, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = many_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in many_nml is something wrong.'
  end if
  write( 6, nml = many_nml )
  !
  close( FileUnit )

  call fi_open( CtlFN4, "read", FileUnit )
  !
  rewind( FileUnit )
  read( FileUnit, nml = mpisubloop_nml, iostat = iostat )
  if ( iostat > 0 ) then
    stop 'Namelist in mpisubloop_nml is something wrong.'
  end if
  write( 6, nml = mpisubloop_nml )
  !
  close( FileUnit )


  ! save
  WaveNumSSave = WaveNumS
  WaveNumESave = WaveNumE


#ifdef MPI
#else
  if ( NProcsMPISubLoop /= 1 ) then
    write( 6, * ) 'Unexpected NProcsMPISubLoop: ', NProcsMPISubLoop
    stop
  end if
#endif
  if ( mod( NProcs, NProcsMPISubLoop ) /= 0 ) then
    write( 6, * ) 'Unexpected NProcsMPISubLoop: ', NProcsMPISubLoop
    write( 6, * ) 'NProcsMPISubLoop has to be a divisor of NProcs'
    stop
  end if


  ! Temporal
!!$  if ( NProcsWaveNumLoop /= 1 ) then
!!$    write( 6, * ) 'NProcsWaveNumLoop has to be 1, now: ', NProcsWaveNumLoop
!!$    stop
!!$  end if


  NTemp    = iTempe    - iTemps    + 1
  NVMRMol1 = iVMRMol1e - iVMRMol1s + 1
  NVMRMol2 = iVMRMol2e - iVMRMol2s + 1
  NVMRMol3 = iVMRMol3e - iVMRMol3s + 1

!!$  NLoop = NTemp * NVMRMol1 * NVMRMol2 * NVMRMol3
  NLoop = NTemp * NVMRMol1 * NVMRMol2 * NVMRMol3 * NProcsMPISubLoop

  iLoope = 0
  do iRank = 0, MyRank
    iLoops = iLoope + 1
    iLoope = iLoops + NLoop / NProcs - 1
    if ( iRank < mod( NLoop, NProcs ) ) then
      iLoope = iLoope + 1
    end if
  end do
  ! check
  if ( iLoope > NLoop ) then
    write( 6, * ) 'Unexpected number of maximum of loop: ', iLoope, NLoop
  end if



  do iLoop = iLoops, iLoope

!!$    iTemp       = iTemps       &
!!$      & + ( iLoop - 1 ) / ( NVMRMol1 * NVMRMol2 * NVMRMol3 )
!!$    iVMRMol1 = iVMRMol1s &
!!$      & + mod( ( iLoop - 1 ) / ( NVMRMol2 * NVMRMol3 ), NVMRMol1 )
!!$    iVMRMol2 = iVMRMol2s &
!!$      & + mod( ( iLoop - 1 ) / ( NVMRMol3 ), NVMRMol2 )
!!$    iVMRMol3 = iVMRMol3s &
!!$      & + mod( ( iLoop - 1 ), NVMRMol3 )

    iTemp       = iTemps       &
      & + ( iLoop - 1 ) / ( NVMRMol1 * NVMRMol2 * NVMRMol3 * NProcsMPISubLoop )
    iVMRMol1 = iVMRMol1s &
      & + mod( ( iLoop - 1 ) / ( NVMRMol2 * NVMRMol3 * NProcsMPISubLoop ), NVMRMol1 )
    iVMRMol2 = iVMRMol2s &
      & + mod( ( iLoop - 1 ) / ( NVMRMol3 * NProcsMPISubLoop ), NVMRMol2 )
    iVMRMol3 = iVMRMol3s &
      & + mod( ( iLoop - 1 ) / ( NProcsMPISubLoop ), NVMRMol3 )
    iProcWaveNumLoop = 0 &
      & + mod( ( iLoop - 1 ), NProcsMPISubLoop )


!!$    write( 6, * ) iLoop, &
!!$      & iTemp, iVMRMol1, iVMRMol2, iVMRMol3, iProcWaveNumLoop


    write( FileNameTail, '(a,i3.3,3(a,i2.2,a,i3.3),a)' ) &
      & "T-", iTemp, &
      & "_M", Mol1Num, "-", iVMRMol1, &
      & "_M", Mol2Num, "-", iVMRMol2, &
      & "_M", Mol3Num, "-", iVMRMol3, &
      & ".nc"


    InFileName     = "data_profile/Profile_" // trim( FileNameTail )
    OutputFileName = trim( OutDir ) // "/ac_" // trim( FileNameTail )

    write( 6, * ) 'Input : ', trim( InFileName )
    write( 6, * ) 'Output: ', trim( OutputFileName )


    PressName    = "Press"
    MolNumName   = "MolNum"
    TempName     = "Temp"
    SurfTempName = "Ts"
    VMRName      = "VMR"



    call IOProfileInqDimLen(      &
      & InFileName, PressName, &
      & NLev &
      & )
    call IOProfileInqDimLen(      &
      & InFileName, MolNumName, &
      & NMolNum &
      & )
    allocate( z_Press   ( NLev    ) )
    allocate( m_MolNum  ( NMolNum ) )
    allocate( z_Temp    ( NLev    ) )
    allocate( zm_VMR    ( NLev, NMolNum ) )
    call IOProfileRead(      &
      & InFileName, PressName, MolNumName, TempName, SurfTempName, VMRName, &
      & NLev, NMolNum, &
      & z_Press, m_MolNum, z_Temp, SurfTemp, zm_VMR     &
      & )
    allocate( z_FlagCalc( NLev ) )
    call IOProfileReadFlag( &
      & InFileName,               &
      & NLev,                     &
      & z_FlagCalc                &
      & )


    allocate( m_FlagCalcLine(NMolNum) )
    allocate( m_FlagCalcCont(NMolNum) )
    m_FlagCalcLine = FlagCalcLine(1:NMolNum)
    m_FlagCalcCont = FlagCalcCont(1:NMolNum)
    write( 6, * ) 'Flags for calculation of lines and continuum:'
    do m = 1, NMolNum
      write( 6, '(a,l,a,l,a,i2)' ) &
        & '  Line = ', m_FlagCalcLine(m), &
        & ', Continuum = ', m_FlagCalcCont(m), &
        & ' for molecule ', m_MolNum(m)
    end do

    allocate( m_LineCutOffWaveNum(NMolNum) )
    m_LineCutOffWaveNum = LineCutOffWaveNum(1:NMolNum)
    write( 6, * ) 'Line cut off wavenumbers:'
    do m = 1, NMolNum
      write( 6, '(a,i4,a,i2.0)' ) &
        & '  Cut-off at ', int( m_LineCutOffWaveNum(m)*1d-2 ), &
        & ' cm-1 for molecule ', m_MolNum(m)
    end do

    allocate( m_AirBroadHalfWidFac(NMolNum) )
    m_AirBroadHalfWidFac = AirBroadHalfWidFac(1:NMolNum)
    write( 6, * ) 'Air-broadening half width factors:'
    do m = 1, NMolNum
      write( 6, '(a,f5.3,a,i2)' ) &
        & '  Air-broadening half width is multiplied by ', m_AirBroadHalfWidFac(m), &
        & ' for molecule ', m_MolNum(m)
    end do


    NHIT = 0
    do n = 1, MaxNmlArySize
      if ( len_trim( HITFileName(n) ) == 0 ) exit
    end do
    NHIT = n - 1
    allocate( a_HITFileName(NHIT) )
    do n = 1, NHIT
      a_HITFileName(n) = HITFileName(n)
    end do
    ! check
    if ( NHIT /= NMolNum ) then
      write( 6, * ) 'Number of HITRAN data files specified in gas_nml does not match to number of species in file, ', &
        trim( InFileName ), '.'
      write( 6, * ) 'Number of HITRAN data files: ', NHIT
      write( 6, * ) 'Number of species          : ', NMolNum
      stop
    end if

    NCIA = 0
    do n = 1, MaxNmlArySize
      if ( len_trim( CIAFileName(n) ) == 0 ) exit
    end do
    NCIA = n - 1
    allocate( a_CIAFileName(NCIA) )
    do n = 1, NCIA
      a_CIAFileName(n) = CIAFileName(n)
    end do

    NXSC = 0
    do n = 1, MaxNmlArySize
      if ( len_trim( XSCFileName(n) ) == 0 ) exit
    end do
    NXSC = n - 1
    allocate( a_XSCFileName(NXSC) )
    do n = 1, NXSC
      a_XSCFileName(n) = XSCFileName(n)
    end do



    ! Modify wavenumber range for a process
    !
    WaveNumS = WaveNumSSave
    WaveNumE = WaveNumESave
    !
    if ( NProcsMPISubLoop == 1 ) then
      ks = 1
      ke = NLev
    else


#ifdef PARALLEL_WAVENUMBER
      write( 6, * ) "Merge wavenumbers"
      !
      ! MPI for wavenumbers
      !
!!$      call ACCalcUtilsModifyWavenumberRange1( &
!!$        & NProcs, MyRank, &
!!$        & DelWaveNum, &
!!$        & WaveNumS, WaveNumE &
!!$        & )
      call ACCalcUtilsModifyWavenumberRange2( &
        & FlagNetCDF, NHIT, a_HITFileName, &
!          & NProcs, MyRank, &
        & NProcsMPISubLoop, iProcWaveNumLoop, &
        & NMolNum, m_MolNum, &
        & DelWaveNum, &
        & WaveNumS, WaveNumE &
        & )
!!$      call ACCalcUtilsModifyWavenumberRange3( &
!!$        & FlagNetCDF, NHIT, a_HITFileName, &
!!$!          & NProcs, MyRank, &
!!$        & NProcsMPISubLoop, iProcWaveNumLoop, &
!!$        & NMolNum, m_MolNum, m_LineCutOffWaveNum, &
!!$        & DelWaveNum, &
!!$        & WaveNumS, WaveNumE &
!!$        & )

      write( OutputFileNameTmp, '(a,a,i6.6,a)' ) &
        & OutputFileName(1:len_trim(OutputFileName)-3), &
        & "_", &
        & FileNumS + iProcWaveNumLoop, &
        & ".nc"
      OutputFileName = OutputFileNameTmp
      write( 6, * ) "Output file is changed to ", trim( OutputFileName ), &
        & ' for wavenumber from ', WaveNumS, ' to ', WaveNumE, ' m-1'
!!$      write( 6, * ) "#", iProcWaveNumLoop, ":", WaveNumS, WaveNumE, trim( OutputFileName )

      ks = 1
      ke = NLev

#else
      write( 6, * ) "Merge vertical domain"
      !
      ! MPI for vertical levels
      !
      if ( NLev < NProcsMPISubLoop ) then
        write( 6, * ) 'Number of processes is too large: ', NLev, NProcsMPISubLoop
        stop
      end if
      !
      ke = 0
      do iRank = 0, iProcWaveNumLoop
        ks = ke + 1
        ke = ks + NLev / NProcsMPISubLoop - 1
        if ( iRank < mod( NLev, NProcsMPISubLoop ) ) then
          ke = ke + 1
        end if
      end do
      ! check
      if ( ke > NLev ) then
        write( 6, * ) 'Unexpected number of maximum of loop: ', ke, NLev
      end if

      write( OutputFileNameTmp, '(a,a,i6.6,a)' ) &
        & OutputFileName(1:len_trim(OutputFileName)-3), &
        & "_", &
        & FileNumS + iProcWaveNumLoop, &
        & ".nc"
      OutputFileName = OutputFileNameTmp
      if ( iProcWaveNumLoop == 0 ) then
        write( 6, * ) "# Vertical domain is divided by MPI."
      end if
      write( 6, * ) "#", iProcWaveNumLoop, ":", ks, ke, trim( OutputFileName )

#endif
    end if



    allocate( zm_VMRShrink( ks:ke, NMolNum ) )

    do m = 1, NMolNum
      do k = ks, ke
        zm_VMRShrink(k,m) = zm_VMR(k,m)
      end do
    end do


    call ACCalc( &
      & DelWaveNum, WaveNumS, WaveNumE, &
      & FlagNetCDF, NHIT, a_HITFileName, &
      & NCIA, a_CIAFileName, &
      & NXSC, a_XSCFileName, &
      & OutputFileName, &
!!$      & NLev, NMolNum, z_Press, m_MolNum, z_Temp, zm_VMR, &
      & ke-ks+1, NMolNum, z_Press(ks:ke), m_MolNum, z_Temp(ks:ke), zm_VMRShrink(ks:ke,:), &
      & m_FlagCalcLine, m_FlagCalcCont, &
      & m_LineCutOffWaveNum, m_AirBroadHalfWidFac, &
      & z_FlagCalc(ks:ke) &
      & )

    deallocate( zm_VMRShrink )

    deallocate( a_HITFileName )
    deallocate( a_CIAFileName )
    deallocate( a_XSCFileName )
    deallocate( z_Press  )
    deallocate( m_MolNum )
    deallocate( z_Temp   )
    deallocate( zm_VMR   )

    deallocate( z_FlagCalc )

    deallocate( m_FlagCalcLine )
    deallocate( m_FlagCalcCont )
    deallocate( m_LineCutOffWaveNum )
    deallocate( m_AirBroadHalfWidFac )

  end do



  write( 6, * ) 'Finish'


#ifdef MPI
  call mpi_finalize( IErr )
#endif

end program main

