module ave_optprop_particle

  use vtype_module

  implicit none

  private

  public :: AveOptPropParticle


  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------

  subroutine AveOptPropParticle( &
    & StrSpeNcFN, PtclOptPropNcFN, OutNcFN, &
    & AtmTemp, &
    & NBand, ab_BandWNBnds &
    & )

    use constants0
    use ni3_module
    use planck_func, only : PF

    character(*), intent(in) :: StrSpeNcFN
    character(*), intent(in) :: PtclOptPropNcFN
    character(*), intent(in) :: OutNcFN
    real(DP)    , intent(in) :: AtmTemp
    integer     , intent(in) :: NBand
    real(DP)    , intent(in) :: ab_BandWNBnds(2,NBand)


    ! Local variables
    !
    character(stdstr)              :: Mode
    integer                        :: NDims
    character(stdstr), allocatable :: a_DimNames( : )
    character(stdstr)              :: Name
    character(stdstr)              :: StdName
    character(extstr)              :: LongName
    character(stdstr)              :: Units
!!$  character(extstr)              :: Comment
    character(extstr*4)            :: CommentStrSpe

    integer, parameter     :: NBnd = 2
    integer                :: a_BndNum(NBnd)

    integer, allocatable :: b_BandNum(:)
    real(8), allocatable :: b_BandWN (:)
    integer, allocatable :: ab_BandWNIndexBnds(:,:)

    real(8), allocatable :: w_StrSpe     (:)

    real(8), allocatable :: w_PF         (:)

    real(8), allocatable :: w_Weight       (:)
    real(8), allocatable :: w_WeightCorrect(:)


    real(8), allocatable :: w_RayScatCoefNonRadAct(:)
    real(8), allocatable :: wm_RayScatCoef(:,:)

    integer(i4b)                   :: NChar
    integer(i4b)     , allocatable :: a_CharNum(:)
    integer                        :: NParticle
    integer                        :: ParticleNRadiusMax
    integer          , allocatable :: a_ParticleNum      (:)
    integer          , allocatable :: r_ParticleRadiusNum(:)
    character(stdstr), allocatable :: a_ParticleName   (:)
    integer          , allocatable :: a_ParticleNRadius(:)
    real(8)          , allocatable :: ra_ParticleRadius(:,:)

    real(8)          , allocatable :: wra_QExt  (:,:,:)
    real(8)          , allocatable :: wra_SSA   (:,:,:)
    real(8)          , allocatable :: wra_AF    (:,:,:)


    real(8)          , allocatable :: b_BandAveRayScatCoefNonRadAct(:)
    real(8)          , allocatable :: bm_BandAveRayScatCoef(:,:)
    real(8)          , allocatable :: bra_BandAveQExt  (:,:,:)
    real(8)          , allocatable :: bra_BandAveSSA   (:,:,:)
    real(8)          , allocatable :: bra_BandAveAF    (:,:,:)


    character(stdstr) :: WaveNumAxisName


    integer              :: NWaveNum
    real(8), allocatable :: w_WaveNum(:)

    integer              :: NMolNum
    integer, allocatable :: m_MolNum(:)

    integer              :: TmpNum

    real(8) :: CommentDWN

!!$  real(8) :: Sum

    integer :: InNcID
    integer :: InNcIDPtcl
    integer :: OutNcID

    logical :: FlagPtclExist

    integer :: i
    integer :: l
    integer :: m
    integer :: n


    a_BndNum(1) = 1
    a_BndNum(2) = 2


    WaveNumAxisName = "WaveNum"

    Mode = "read"
    call ni3_open( PtclOptPropNcFN, Mode, InNcID )
    call ni3_inq_dimlen( InNcID, WaveNumAxisName, NWaveNum )
    allocate( w_WaveNum    (NWaveNum) )
    call ni3_get_var( InNcID, WaveNumAxisName, w_WaveNum )
    call ni3_inq_dimlen( InNcID, 'MolNum', NMolNum )
    allocate( m_MolNum    (NMolNum) )
    call ni3_get_var( InNcID, 'MolNum', m_MolNum )
    call ni3_close( InNcID )


    allocate( w_RayScatCoefNonRadAct(NWaveNum) )
    allocate( wm_RayScatCoef (NWaveNum,NMolNum) )
    allocate( w_StrSpe       (NWaveNum) )
    allocate( w_PF           (NWaveNum) )
    allocate( w_Weight       (NWaveNum) )
    allocate( w_WeightCorrect(NWaveNum) )


    Mode = "read"
    call ni3_open( PtclOptPropNcFN, Mode, InNcIDPtcl )
    !
    call ni3_get_att( InNcIDPtcl, "_global_", "comment_dwn_m-1", CommentDWN )
    !
    call ni3_inq_dimlen( InNcIDPtcl, WaveNumAxisName, TmpNum )
    if ( TmpNum /= NWaveNum ) then
      write( 6, * ) 'Array size for coefficients of particles is inappropriate: '
      write( 6, * ) '  ', trim( PtclOptPropNcFN ), ': ', TmpNum
      write( 6, * ) '  ', trim( "Original "     ), ': ', NWaveNum
      stop
    end if
    !
    call ni3_get_var( InNcIDPtcl, "RayScatCoefNonRadAct", w_RayScatCoefNonRadAct )
    call ni3_get_var( InNcIDPtcl, "RayScatCoef"         , wm_RayScatCoef )
    !
    FlagPtclExist = ni3_chk_var( InNcIDPtcl, "CharNum" )
    !
    if ( FlagPtclExist ) then
      call ni3_inq_dimlen( InNcIDPtcl, "CharNum"      , NChar              )
      call ni3_inq_dimlen( InNcIDPtcl, "PtclNum"      , NParticle          )
      call ni3_inq_dimlen( InNcIDPtcl, "PtclRadiusNum", ParticleNRadiusMax )
    else
      NChar              = 0
      NParticle          = 0
      ParticleNRadiusMax = 0
    end if
    allocate( a_CharNum          (NChar             ) )
    allocate( a_ParticleNum      (NParticle         ) )
    allocate( r_ParticleRadiusNum(ParticleNRadiusMax) )
    allocate( a_ParticleName     (NParticle         ) )
    allocate( a_ParticleNRadius  (NParticle         ) )
    allocate( ra_ParticleRadius(ParticleNRadiusMax,NParticle) )
    allocate( wra_QExt         (NWaveNum,ParticleNRadiusMax,NParticle) )
    allocate( wra_SSA          (NWaveNum,ParticleNRadiusMax,NParticle) )
    allocate( wra_AF           (NWaveNum,ParticleNRadiusMax,NParticle) )
    if ( FlagPtclExist ) then
      call ni3_get_var( InNcIDPtcl, "CharNum"      , a_CharNum           )
      call ni3_get_var( InNcIDPtcl, "PtclNum"      , a_ParticleNum       )
      call ni3_get_var( InNcIDPtcl, "PtclRadiusNum", r_ParticleRadiusNum )
      call ni3_get_var( InNcIDPtcl, "PtclName"     , a_ParticleName      )
      call ni3_get_var( InNcIDPtcl, "PtclNRadius"  , a_ParticleNRadius   )
      call ni3_get_var( InNcIDPtcl, "PtclRadius"   , ra_ParticleRadius   )
      call ni3_get_var( InNcIDPtcl, "QExt"         , wra_QExt            )
      call ni3_get_var( InNcIDPtcl, "SSA"          , wra_SSA             )
      call ni3_get_var( InNcIDPtcl, "AF"           , wra_AF              )
    end if


    Mode = "read"
    call ni3_open( StrSpeNcFN, Mode, InNcID )
    call ni3_get_att( InNcID, '_global_', 'comment_strspe', CommentStrSpe )
    call ni3_inq_dimlen( InNcID, WaveNumAxisName, TmpNum )
    if ( TmpNum /= NWaveNum ) then
      write( 6, * ) 'Array size for stellar spectrum is inappropriate: '
      write( 6, * ) '  ', trim( StrSpeNcFN      ), ': ', TmpNum
      write( 6, * ) '  ', trim( PtclOptPropNcFN ), ': ', NWaveNum
      stop
    end if
    call ni3_get_var( InNcID, 'InFlux', w_StrSpe )
    call ni3_close( InNcID )


    do l = 1, NWaveNum
      w_PF(l) = PI * PF( w_WaveNum(l), AtmTemp )
    end do


    ! calculation of weight
    w_Weight = w_StrSpe + w_PF


    allocate( b_BandNum         (NBand) )
    allocate( ab_BandWNIndexBnds(2,NBand) )


    do n = 1, NBand
      b_BandNum(n) = n
    end do


    ! Look for bound index
    do l = 1, NWaveNum
      if ( w_WaveNum(l) >= ab_BandWNBnds(1,1) ) exit
    end do
    ab_BandWNIndexBnds(1,1) = l
    !
    do n = 1, NBand
      do l = ab_BandWNIndexBnds(1,n), NWaveNum
        if ( w_WaveNum(l) > ab_BandWNBnds(2,n) ) exit
      end do
      ab_BandWNIndexBnds(2,n) = l-1
      if ( n < NBand ) then
        ab_BandWNIndexBnds(1,n+1) = ab_BandWNIndexBnds(2,n) + 1
      end if
    end do


    ! check band
    do n = 1, NBand
      if ( ab_BandWNIndexBnds(2,n) - ab_BandWNIndexBnds(1,n) + 1 <= 0 ) then
        write( 6, * ) 'ERROR: Band ', n, ' has no absorption coefficient.'
        stop
      end if
    end do


    ! calculate band average wavenumber
    allocate( b_BandWN (NBand) )
    do n = 1, NBand
      b_BandWN(n) = 0.0d0
      do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
        b_BandWN(n) = b_BandWN(n) + w_WaveNum(l)
      end do
      if ( ab_BandWNIndexBnds(2,n)-ab_BandWNIndexBnds(1,n) >= 0 ) then
        b_BandWN(n) = b_BandWN(n) &
          & / ( ab_BandWNIndexBnds(2,n) - ab_BandWNIndexBnds(1,n) + 1 )
      end if
    end do


    allocate( b_BandAveRayScatCoefNonRadAct(NBand) )
    allocate( bm_BandAveRayScatCoef(NBand,NMolNum) )

    call BandAverage( &
      & NWaveNum, w_RayScatCoefNonRadAct, w_Weight, &
      & NBand, ab_BandWNIndexBnds, &
      & b_BandAveRayScatCoefNonRadAct &
      & )
    do m = 1, NMolNum
      call BandAverage( &
        & NWaveNum, wm_RayScatCoef(:,m), w_Weight, &
        & NBand, ab_BandWNIndexBnds, &
        & bm_BandAveRayScatCoef(:,m) &
        & )
    end do

    allocate( bra_BandAveQExt( NBand, ParticleNRadiusMax, NParticle ) )
    allocate( bra_BandAveSSA ( NBand, ParticleNRadiusMax, NParticle ) )
    allocate( bra_BandAveAF  ( NBand, ParticleNRadiusMax, NParticle ) )

    do i = 1, NParticle
      do m = 1, a_ParticleNRadius(i)
        call BandAverage( &
          & NWaveNum, wra_QExt(:,m,i), w_Weight, &
          & NBand, ab_BandWNIndexBnds, &
          & bra_BandAveQExt(:,m,i) &
          & )
        w_WeightCorrect = w_Weight * wra_QExt(:,m,i)
        call BandAverage( &
          & NWaveNum, wra_SSA (:,m,i), w_WeightCorrect, &
          & NBand, ab_BandWNIndexBnds, &
          & bra_BandAveSSA (:,m,i) &
          & )
        w_WeightCorrect = w_Weight * wra_QExt(:,m,i) * wra_SSA(:,m,i)
        call BandAverage( &
          & NWaveNum, wra_AF  (:,m,i), w_WeightCorrect, &
          & NBand, ab_BandWNIndexBnds, &
          & bra_BandAveAF  (:,m,i) &
          & )
      end do
      do m = a_ParticleNRadius(i)+1, ParticleNRadiusMax
        bra_BandAveQExt(:,m,i) = -1.0d100
        bra_BandAveSSA (:,m,i) = -1.0d100
        bra_BandAveAF  (:,m,i) = -1.0d100
      end do
    end do


    ! Output

    Mode = "new"
    call ni3_open( OutNcFN, Mode, OutNcID )

    call ni3_put_att( OutNcID, "_global_", "comment_dwn_m-1", CommentDWN )
    call ni3_put_att( OutNcID, '_global_', 'comment_strspe', CommentStrSpe )

    Name     = "BndNum"
    StdName  = "BndNum"
    LongName = "bound number"
    Units    = "1"
    call ni3_set_dim( OutNcID, Name, NI3_INT   , a_BndNum, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "BandNum"
    StdName  = "BandNum"
    LongName = "band number"
    Units    = "1"
    call ni3_set_dim( OutNcID, Name, NI3_INT, b_BandNum, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "BandWaveNum"
    StdName  = "BandWaveNum"
    LongName = "band wavenumber"
    Units    = "m-1"
    call ni3_set_dim( OutNcID, Name, NI3_DOUBLE, b_BandWN, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    name    = "MolNum"
    call ni3_set_dim( OutNcID, name, NI3_INT, m_MolNum )
    call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
    !
    if ( FlagPtclExist ) then
      name    = "CharNum"
      call ni3_set_dim( OutNcID, name, NI3_INT, a_CharNum )
      call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
      !
      Name     = "PtclNum"
      call ni3_set_dim( OutNcID, Name, NI3_INT, a_ParticleNum )
      call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
      !
      Name     = "PtclRadiusNum"
      call ni3_set_dim( OutNcID, Name, NI3_INT, r_ParticleRadiusNum )
      call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
      !
      !
      !
      NDims = 2
      allocate( a_DimNames( NDims ) )
      !
      a_DimNames( 1 ) = "CharNum"
      a_DimNames( 2 ) = "PtclNum"
      Name     = "PtclName"
      call ni3_def_var( OutNcID, Name, NI3_CHAR, NDims, a_DimNames )
      call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
      call ni3_put_var( OutNcID, Name, a_ParticleName )
      !
      deallocate( a_DimNames )
      !
      !
      !
      NDims = 1
      allocate( a_DimNames( NDims ) )
      !
      a_DimNames( 1 ) = "PtclNum"
      Name     = "PtclNRadius"
      call ni3_def_var( OutNcID, Name, NI3_INT, NDims, a_DimNames )
      call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
      call ni3_put_var( OutNcID, Name, a_ParticleNRadius )
      !
      deallocate( a_DimNames )
      !
      !
      NDims = 2
      allocate( a_DimNames( NDims ) )
      !
      a_DimNames( 1 ) = "PtclRadiusNum"
      a_DimNames( 2 ) = "PtclNum"
      Name     = "PtclRadius"
      StdName  = Name
      LongName = "particle radius"
      Units    = "m"
      call ni3_def_var( OutNcID, name, NI3_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
      call ni3_cp_atts( InNcIDPtcl, OutNcID, Name, Name )
      call ni3_put_var( OutNcID, name, ra_ParticleRadius )
      !
      deallocate( a_DimNames )
    end if
    !
    !
    NDims = 2
    allocate( a_DimNames( NDims ) )
    !
    a_DimNames( 1 ) = "BndNum"
    a_DimNames( 2 ) = "BandNum"
    Name     = "BandWaveNumBnds"
    StdName  = Name
    LongName = "band wavenumber bound"
    Units    = "m-1"
    call ni3_def_var( OutNcID, name, NI3_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( OutNcID, name, ab_BandWNBnds )
    !
    deallocate( a_DimNames )


    NDims = 1
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BandNum"
!!$  a_DimNames( 1 ) = "BandWaveNum"
    Name     = "BandAveRayScatCoefNonRadAct"
    StdName  = Name
    LongName = "Band averaged Rayleigh scattering coefficient for non-radiative active molecule"
    Units    = "(molecule m-2)-1"
    call ni3_def_var( OutNcID, Name, NI3_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( OutNcID, Name, b_BandAveRayScatCoefNonRadAct )
    deallocate( a_DimNames )

    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BandNum"
!!$  a_DimNames( 1 ) = "BandWaveNum"
    a_DimNames( 2 ) = "MolNum"
    Name     = "BandAveRayScatCoef"
    StdName  = Name
    LongName = "Band averaged Rayleigh scattering coefficient"
    Units    = "(molecule m-2)-1"
    call ni3_def_var( OutNcID, Name, NI3_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( OutNcID, Name, bm_BandAveRayScatCoef )
    deallocate( a_DimNames )

    if ( FlagPtclExist ) then
      NDims = 3
      allocate( a_DimNames( NDims ) )
      !
      a_DimNames( 1 ) = "BandNum"
!!$  a_DimNames( 1 ) = "BandWaveNum"
      a_DimNames( 2 ) = "PtclRadiusNum"
      a_DimNames( 3 ) = "PtclNum"
      !
      Name     = "BandAveQExt"
      StdName  = Name
      LongName = "Band averaged extinction efficiency factor of particles"
      Units   = "1"
      call ni3_def_var( OutNcID, Name, NI3_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
!!$  call ni3_cp_atts( InNcIDPtcl, OutNcID, "QExt", Name )
      call ni3_put_var( OutNcID, Name, bra_BandAveQExt )
      !
      Name     = "BandAveSSA"
      StdName  = Name
      LongName = "Band averaged single scattering albedo of particles"
      Units    = "1"
      call ni3_def_var( OutNcID, Name, NI3_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
!!$  call ni3_cp_atts( InNcIDPtcl, OutNcID, "SSA", Name )
      call ni3_put_var( OutNcID, Name, bra_BandAveSSA  )
      !
      Name     = "BandAveAF"
      StdName  = Name
      LongName = "Band averaged asymmetry factor of particles"
      Units    = "1"
      call ni3_def_var( OutNcID, Name, NI3_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
!!$  call ni3_cp_atts( InNcIDPtcl, OutNcID, "AF", Name )
      call ni3_put_var( OutNcID, Name, bra_BandAveAF   )
    end if

    call ni3_close( InNcIDPtcl )

    call ni3_close( OutNcID )


    deallocate( b_BandAveRayScatCoefNonRadAct )
    deallocate( bm_BandAveRayScatCoef )

    deallocate( w_RayScatCoefNonRadAct )
    deallocate( wm_RayScatCoef  )
    deallocate( w_Weight        )
    deallocate( w_WeightCorrect )


    deallocate( b_BandWN           )
    deallocate( ab_BandWNIndexBnds )

    deallocate( w_WaveNum )


  contains

    !--------------------------------------------------------------------------

    subroutine BandAverage( &
      & NWaveNum, w_Array, w_Weight, &
      & NBand, ab_BandWNIndexBnds, &
      & b_BandAveArray &
      & )

      integer, intent(in ) :: NWaveNum
      real(8), intent(in ) :: w_Array (NWaveNum)
      real(8), intent(in ) :: w_Weight(NWaveNum)
      integer, intent(in ) :: NBand
      integer, intent(in ) :: ab_BandWNIndexBnds(2,NBand)
      real(8), intent(out) :: b_BandAveArray(NBand)


      ! Local variables
      !
      real(8) :: Sum

      integer :: l
      integer :: n


      do n = 1, NBand
        b_BandAveArray(n) = 0.0d0
        Sum = 0.0d0
        do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
          b_BandAveArray(n) = b_BandAveArray(n) &
            & + w_Array(l) * w_Weight(l)
          Sum = Sum + w_Weight(l)
        end do
        if ( ab_BandWNIndexBnds(2,n)-ab_BandWNIndexBnds(1,n) >= 0 ) then
          b_BandAveArray(n) = b_BandAveArray(n) / Sum
        else
          b_BandAveArray(n) = 0.0d0
        end if
      end do


    end subroutine BandAverage

    !--------------------------------------------------------------------------

  end subroutine AveOptPropParticle

  !----------------------------------------------------------------------------

end module ave_optprop_particle
