module sortac_module

  use vtype_module

  implicit none

  private

  public :: sortac

  character(stdstr), save :: AbsCoefVarName = 'AbsCoef'

  character(128), save :: ModuleName = 'sortac_module'


contains

  !----------------------------------------------------------------------------

  subroutine SortAC( &
    & FlagSort, &
!!$    & InNcFN, OutNcFN, &
    & InNcFN, InNcFNTemp, OutNcFN, &
    & FlagChkConstVMR, FlagBinning, FlagBinningDemonstration, &
    & FlagKeyMolACAverage, FlagNonKeyMolACAverage, &
    & FlagPFRatioOne, &
    & FlagSortPFRatioAverage, &
    & FlagCorrelation, &
    & FlagKeyMolAbsOverlap, &
    & StrSpeNcFN, &
    & NBand, ab_BandWNBnds, b_ArgBandNBin, &
    & NKeyMol, ab_BandKeyMolNum, &
    & NGaussReg, a_GaussRegBnds, cb_BandGaussRegNBin, &
    & PressForPFSort, &
    & FlagCopyVarMolAbsCoef, VarMolRangeNcFN, m_VarMolNum, m_VMRVarMol &
    & )

    use ni3_module
    use netcdf

    use sort_module, only : sort_heap
    use rad_kd_utils, only : InqPFIntGQNum

    use planck_func_wrapper, only : Integ_PF_GQ_Array1D

    use constants0, only: PI

    logical     , intent(in ) :: FlagSort
    character(*), intent(in ) :: InNcFN
    character(*), intent(in ) :: InNcFNTemp
    character(*), intent(in ) :: OutNcFN

    logical, intent(in ) :: FlagChkConstVMR
    logical, intent(in ) :: FlagBinning
    logical, intent(in ) :: FlagBinningDemonstration
    logical, intent(in ) :: FlagKeyMolACAverage
    logical, intent(in ) :: FlagNonKeyMolACAverage
    logical, intent(in ) :: FlagPFRatioOne
    logical, intent(in ) :: FlagSortPFRatioAverage
    logical, intent(in ) :: FlagCorrelation
    logical, intent(in ) :: FlagKeyMolAbsOverlap
    character(len=*), intent(in ) :: StrSpeNcFN
    integer, intent(in ) :: NBand
    real(8), intent(in ) :: ab_BandWNBnds(2,NBand)
                                    ! Wavenumber bounds of bands
    integer, intent(in ) :: b_ArgBandNBin  (1:NBand)
                           ! Number of bins in bands
    integer, intent(in ) :: NKeyMol
    integer, intent(in ) :: ab_BandKeyMolNum(1:NKeyMol,1:NBand)
    integer, intent(in ) :: NGaussReg
    real(8), intent(in ) :: a_GaussRegBnds(0:NGaussReg)
    integer, intent(in ) :: cb_BandGaussRegNBin(NGaussReg,NBand)
    real(DP), intent(in ) :: PressForPFSort
    logical     , intent(in), optional :: FlagCopyVarMolAbsCoef
    character(*), intent(in), optional :: VarMolRangeNcFN
    integer     , intent(in), optional :: m_VarMolNum(3)
    real(8)     , intent(in), optional :: m_VMRVarMol(3)

    ! local variables
    !
    character(stdstr) :: Mode
    integer           :: NcID
    integer           :: NDims
    character(stdstr), allocatable :: a_DimNames( : )
    character(stdstr) :: Name
    character(stdstr) :: StdName
    character(stdstr) :: LongName
    character(stdstr) :: Units
    character(extstr) :: Comment
    character(extstr*4) :: CommentStrSpe
    real(DP)            :: CommentDWN


    integer                :: b_BandNBin  (1:NBand)

    integer                :: NPress
    real(8), allocatable   :: p_Press(:)
    real(8), allocatable   :: p_Temp (:)

    integer                :: NPressForPF
    real(8), allocatable   :: r_PressForPF(:)
    real(8), allocatable   :: r_TempForPF(:)

    integer                :: NWaveNum
    real(8), allocatable   :: w_WaveNum(:)

    integer                :: NMol
    integer, allocatable   :: m_MolNum(:)             ! Molecular number

    real(8), allocatable   :: pmw_AbsCoef(:,:,:)     ! Absorption coefficient
    real(8), allocatable   :: wm_AbsCoef(:,:)
    real(8), allocatable   :: rw_AtmPFRatio(:,:)       ! Planck func. ratio
    real(8), allocatable   :: rw_StrPFRatio(:,:)       ! Planck func. ratio
    real(8), allocatable   :: w_AtmPFRatio(:)
    real(8), allocatable   :: w_StrPFRatio(:)
!!$    real(8), allocatable   :: w_AbsCoef(:)
    integer, allocatable   :: w_SortedIndex   (:)
!!$    integer, allocatable   :: w_SortedIndexTmp(:)
    integer, allocatable   :: wp_SortedIndexSave(:,:)

    real(8), allocatable   :: pm_VMR(:,:)

    integer :: b_NKeyMol(NBand)
    integer :: MinKeyMol
    integer :: b_IndexFirstKeyMol(NBand)

    real(8), allocatable   :: pw_AbsCoefSum(:,:)

    real(8), allocatable   :: w_StrSpe(:)



!!$  real(8) :: PressMin
!!$  real(8) :: PressMax
!!$  integer :: NPressPer10
!!$
!!$  real(8) :: TempMin
!!$  real(8) :: TempMax
!!$  real(8) :: TempInc

!!$  real(8) :: Press
!!$  real(8) :: Temp

    integer              :: a_BndNum(2)

    integer, allocatable :: b_BandNum(:)
    real(8), allocatable :: b_BandWN(:)

    integer, allocatable :: ab_BandWNIndexBnds(:,:)
                           ! Wavenumber index bounds of bands
    integer, allocatable :: ab_BandBinIndexBnds(:,:)
                           ! Index bounds of bins in bands

    integer              :: NBin
                           ! Number of bins
    integer, allocatable :: c_BinNum    (:)
                           ! Bin number
    real(8), allocatable :: c_BinPointWN(:)
                           ! Bin point wavenumber
    real(8), allocatable :: c_BinWeight (:)
                           ! Bin weight
    integer, allocatable :: ac_BinWNIndexBnds(:,:)
                           ! Wavenumber index bounds of bins
    real(8), allocatable :: pmc_BinAbsCoef(:,:,:)     ! Absorption coefficient
    real(8), allocatable :: rc_BinAtmPFRatio (:,:)    ! Planck func. ratio
    real(8), allocatable :: rc_BinStrPFRatio (:,:)    ! Planck func. ratio
    real(8), allocatable :: c_BinStrPFRatio  (:)      ! Planck func. ratio

!!$    real(8), allocatable :: c_BinWNBnds(:)

    ! variables for calculating average Planck function in each bin
    integer               :: TmpNBin
    integer               :: b_TmpBandNBin  (1:NBand)
    integer , allocatable :: c_TmpBinNum    (:)
    integer , allocatable :: ab_TmpBandBinIndexBnds(:,:)
    real(DP), allocatable :: c_TmpBinWeight (:)
    real(DP), allocatable :: c_TmpBinPointWN(:)
    integer , allocatable :: ac_TmpBinWNIndexBnds(:,:)

    real(8), allocatable :: pmc_TmpBinAbsCoef  (:,:,:)
    real(8), allocatable :: rc_TmpBinAtmPFRatio(:,:)
    real(8), allocatable :: rc_TmpBinStrPFRatio(:,:)
    integer, allocatable :: mc_TmpIDAbsCoefType(:,:)

    real(DP) :: TotalFluxTOA

!!$    real(8), allocatable :: r_DelMass(:)
!!$    real(8) :: DelAbsMass
    real(8) :: Sum

!!$    integer :: kp
!!$    integer :: kf

    real(DP) :: DelWaveNum
    real(DP) :: WNs
    real(DP) :: WNe
    integer  :: GQNum

    integer  :: TmpNum

    real(DP), allocatable :: rb_BandAveAtmPF(:,:)
    real(DP), allocatable :: rb_BandAveStrPF(:,:)
    real(DP), allocatable :: rb_RatioBandIntFluxtoTotFlux(:,:)

    ! mb_IDAbsCoefType / mc_IDAbsCoefType
    !   = -1 : zero abs. coef.
    !   = 0  : abs. coef. per unit abs. molecule
    !   = 1  : abs. coef. per unit atm. molecule
    integer, allocatable :: mb_IDAbsCoefType(:,:)
    integer, allocatable :: mw_IDAbsCoefType(:,:)
    integer, allocatable :: mc_IDAbsCoefType(:,:)

    logical               :: FlagMissingValue
    real(DP)              :: AbsCoefMissingValue
    logical , allocatable :: p_FlagCalc (:)
    real(DP)              :: PFRatioMissingValue


!!$    real(DP) :: TempMin

    integer :: j
    integer :: k
    integer :: l
    integer :: m
    integer :: n
    integer :: iKeyMol

    integer :: ls
    integer :: le

    integer :: ks
    integer :: ke

    integer  :: kIndex
!!$    integer  :: kIndexForPFSort

    integer :: mVarMol


    a_BndNum(1) = 1
    a_BndNum(2) = 2

    allocate( ab_BandWNIndexBnds(2,NBand) )

    allocate( b_BandNum(NBand) )
    do n = 1, NBand
      b_BandNum(n) = n
    end do


    Mode = "read"
    call ni3_open( InNcFN, Mode, NcID )
    !
    call ni3_get_att( NcID, '_global_', 'comment_dwn_m-1', CommentDWN )
    !
    call ni3_inq_dimlen( NcID, "WaveNum", NWaveNum )
    call ni3_inq_dimlen( NcID, "MolNum" , NMol     )
    call ni3_inq_dimlen( NcID, "Press"  , NPress   )
    !
    allocate( w_WaveNum(NWaveNum) )
    allocate( m_MolNum(NMol) )
    allocate( p_Press(NPress) )
    allocate( p_Temp (NPress) )
    !
    allocate( pm_VMR       (NPress, NMol) )
    allocate( pmw_AbsCoef  (NPress, NMol, NWaveNum) )
    !
    call ni3_get_var( NcID, "Press"  , p_Press      )
    call ni3_get_var( NcID, "MolNum" , m_MolNum     )
    call ni3_get_var( NcID, "WaveNum", w_WaveNum    )
    call ni3_get_var( NcID, "Temp"   , p_Temp       )
    call ni3_get_var( NcID, "VMR"    , pm_VMR       )
    call ni3_get_var( NcID, AbsCoefVarName, pmw_AbsCoef )
    allocate( p_FlagCalc(NPress) )
    if ( ni3_chk_att( NcID, AbsCoefVarName, "missing_value" ) ) then
      FlagMissingValue = .true.
      call ni3_get_att( NcID, AbsCoefVarName, "missing_value", AbsCoefMissingValue )
      do k = 1, NPress
        if ( pmw_AbsCoef(k,1,1) == AbsCoefMissingValue ) then
          p_FlagCalc(k) = .false.
        else
          p_FlagCalc(k) = .true.
        end if
      end do
      PFRatioMissingValue = AbsCoefMissingValue
    else
      FlagMissingValue = .false.
      p_FlagCalc = .true.
    end if
    !
    call ni3_close( NcID )


    if ( present( FlagCopyVarMolAbsCoef ) ) then
      if ( ( .not. present( VarMolRangeNcFN ) ) .or. &
        &  ( .not. present( m_VarMolNum     ) ) .or. &
        &  ( .not. present( m_VMRVarMol     ) ) ) then
        write( 6, * ) &
          & 'ERROR: Arguments, VarMolRangeNcFN, m_VarMolNum, m_VMRVarMol, ' &
          & // 'have to be present, if FlagCopyVarMolAbsCoef is present.'
        stop
      end if
      if ( FlagCopyVarMolAbsCoef ) then
        do mVarMol = 1, 3
          do m = 1, NMol
            if ( m_MolNum(m) == m_VarMolNum(mVarMol) ) exit
          end do
          if ( m > NMol ) then
            write( 6, * ) &
              & 'ERROR: Specifiec molecular number for variable molecule cannot be found.'
            write( 6, * ) m_VarMolNum(mVarMol)
            write( 6, * ) m_MolNum
            stop
          end if
          write( 6, * ) &
            & 'Volume mixing ratio for molecule number ', m_MolNum(m), &
            & ' is changed to ', m_VMRVarMol(mVarMol), '.'
          pm_VMR(:,m) = m_VMRVarMol(mVarMol)
        end do
      end if
    end if


    Mode = "read"
    call ni3_open( InNcFNTemp, Mode, NcID )
    !
    call ni3_inq_dimlen( NcID, "Press", NPressForPF )
    if ( ( NPressForPF /= NPress ) .and. ( NPressForPF /= NPress+1 ) ) then
      write( 6, * ) 'In ', trim( ModuleName )
      write( 6, * ) '   Number of levels for temperature is inappropriate.'
      write( 6, * ) '   ', NPressForPF, NPress
      stop
    end if
    !
    allocate( r_PressForPF(NPressForPF) )
    allocate( r_TempForPF (NPressForPF) )
    !
    call ni3_get_var( NcID, "Press"  , r_PressForPF )
    call ni3_get_var( NcID, "Temp"   , r_TempForPF  )
    !
    call ni3_close( NcID )


    allocate( rw_AtmPFRatio(NPressForPF, NWaveNum) )
    allocate( rw_StrPFRatio(NPressForPF, NWaveNum) )


    ! Find closest level for PressForPFSort
!!$    kIndexForPFSort = 1
!!$    do k = 1, NPress
!!$      if (   abs( PressForPFSort - p_Press(k)               ) &
!!$        &  < abs( PressForPFSort - p_Press(kIndexForPFSort) ) ) then
!!$        kIndexForPFSort = k
!!$      end if
!!$    end do
!!$    write( 6, * ) &
!!$      & 'Pressure for Planck function sorting : ', &
!!$      & p_Press(kIndexForPFSort), ' Pa.'
!!$    write( 6, * ) &
!!$      & '  Specified pressure                 : ', &
!!$      & PressForPFSort, ' Pa.'
    write( 6, * ) &
      & 'NOTE (for developer): Pressure for Planck function sorting is not used, now.'



    ! NKeyMol for each band
    do n = 1, NBand
      do iKeyMol = 1, NKeyMol
        if ( ab_BandKeyMolNum(iKeyMol,n) <= 0 ) then
          exit
        end if
      end do
      b_NKeyMol(n) = iKeyMol-1
    end do

    ! Find minimum of key molecular number
    do n = 1, NBand
      iKeyMol = 1
      MinKeyMol = ab_BandKeyMolNum(iKeyMol,n)
      do iKeyMol = 1+1, b_NKeyMol(n)
        MinKeyMol = min( MinKeyMol, ab_BandKeyMolNum(iKeyMol,n) )
      end do
      do m = 1, NMol
        if ( m_MolNum(m) == MinKeyMol ) exit
      end do
      if ( m > NMol ) then
        write( 6, * ) 'Unable to find index for a key molecule.'
        write( 6, * ) m_MolNum
        write( 6, * ) MinKeyMol
        write( 6, * ) ab_BandKeyMolNum
        stop
      end if
      b_IndexFirstKeyMol(n) = m
    end do



    allocate( w_StrSpe( 1:NWaveNum ) )

    Mode = "read"
    call ni3_open( StrSpeNcFN, Mode, NcID )
    call ni3_inq_dimlen( NcID, "WaveNum", TmpNum )
    if ( TmpNum /= NWaveNum ) then
      write( 6, * ) 'Array size for stellar spectrum is inappropriate: ', &
        TmpNum, NWaveNum
      stop
    end if
    call ni3_get_var( NcID, "InFlux", w_StrSpe )
    call ni3_get_att( NcID, '_global_', 'comment_strspe', CommentStrSpe )
    call ni3_close( NcID )


    allocate( mb_IDAbsCoefType(NMol, NBand   ) )
    allocate( mw_IDAbsCoefType(NMol, NWaveNum) )


    if ( FlagChkConstVMR ) then
      do m = 1, NMol
        do k = 1+1, NPress
          if ( pm_VMR(1,m) /= pm_VMR(k,m) ) then
            write( 6, * ) 'ERROR: Mixing ratio is not constant over pressure.'
            write( 6, * ) m, k, pm_VMR(k,m), pm_VMR(1,m)
            stop
          end if
        end do
      end do
    end if




    ! Look for bound index
    loop_search_bound_index_1 : do l = 1, NWaveNum
      if ( w_WaveNum(l) >= ab_BandWNBnds(1,1) ) exit loop_search_bound_index_1
    end do loop_search_bound_index_1
    ab_BandWNIndexBnds(1,1) = l
    !
    do n = 1, NBand
      loop_search_bound_index : do l = ab_BandWNIndexBnds(1,n), NWaveNum
        if ( w_WaveNum(l) > ab_BandWNBnds(2,n) ) exit loop_search_bound_index
      end do loop_search_bound_index
      ab_BandWNIndexBnds(2,n) = l-1
      if ( n < NBand ) then
        ab_BandWNIndexBnds(1,n+1) = ab_BandWNIndexBnds(2,n) + 1
      end if
    end do


    ! check band
    do n = 1, NBand
      if ( ab_BandWNIndexBnds(2,n) - ab_BandWNIndexBnds(1,n) + 1 <= 0 ) then
        write( 6, * ) 'ERROR: Band ', n, ' has no absorption coefficient.'
        stop
      end if
    end do


    ! calculate band average wavenumber
    allocate( b_BandWN (NBand) )
    do n = 1, NBand
!!$      b_BandWN(n) = 0.0d0
!!$      do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
!!$        b_BandWN(n) = b_BandWN(n) + w_WaveNum(l)
!!$      end do
!!$      if ( ab_BandWNIndexBnds(2,n)-ab_BandWNIndexBnds(1,n) >= 0 ) then
!!$        b_BandWN(n) = b_BandWN(n) &
!!$          & / ( ab_BandWNIndexBnds(2,n) - ab_BandWNIndexBnds(1,n) + 1 )
!!$      end if
      !
      b_BandWN(n) = ( ab_BandWNBnds(1,n) + ab_BandWNBnds(2,n) ) / 2.0d0
    end do


    ! Prepare wavenumber integrated band Planck function
!!$    allocate( pb_BandAveAtmPF(NPress,NBand) )
!!$    allocate( pb_BandAveStrPF(NPress,NBand) )
!!$    do n = 1, NBand
!!$      WNs = ab_BandWNBnds(1,n)
!!$      WNe = ab_BandWNBnds(2,n)
!!$      GQNum = InqPFIntGQNum( WNe - WNs )
!!$      call Integ_PF_GQ_Array1D( &
!!$        & WNs, WNe, GQNum, &
!!$        & 1, NPress, &
!!$        & p_Temp, &
!!$        & pb_BandAveAtmPF(:,n) &
!!$        & )
!!$      pb_BandAveAtmPF(:,n) = PI * pb_BandAveAtmPF(:,n) / ( WNe - WNs )
!!$    end do
    !
    allocate( rb_BandAveAtmPF(NPressForPF,NBand) )
    allocate( rb_BandAveStrPF(NPressForPF,NBand) )
    do n = 1, NBand
      WNs = ab_BandWNBnds(1,n)
      WNe = ab_BandWNBnds(2,n)
      GQNum = InqPFIntGQNum( WNe - WNs )
      call Integ_PF_GQ_Array1D( &
        & WNs, WNe, GQNum, &
        & 1, NPressForPF, &
        & r_TempForPF, &
        & rb_BandAveAtmPF(:,n) &
        & )
      rb_BandAveAtmPF(:,n) = PI * rb_BandAveAtmPF(:,n) / ( WNe - WNs )
    end do
    !
    do n = 1, NBand
      WNs = ab_BandWNBnds(1,n)
      WNe = ab_BandWNBnds(2,n)
      DelWaveNum = w_WaveNum(2) - w_WaveNum(1)
      rb_BandAveStrPF(1,n) = 0.0d0
      do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
        rb_BandAveStrPF(1,n) = rb_BandAveStrPF(1,n) + w_StrSpe(l) * DelWaveNum
      end do
      rb_BandAveStrPF(1,n) = rb_BandAveStrPF(1,n) / ( WNe - WNs )
      do k = 1+1, NPressForPF
        rb_BandAveStrPF(k,n) = rb_BandAveStrPF(1,n)
      end do
    end do
    allocate( rb_RatioBandIntFluxtoTotFlux(NPressForPF,NBand) )
    DelWaveNum = w_WaveNum(2) - w_WaveNum(1)
    TotalFluxTOA = 0.0d0
    do l = 1, NWaveNum
      TotalFluxTOA = TotalFluxTOA + w_StrSpe(l) * DelWaveNum
    end do
    do n = 1, NBand
      rb_RatioBandIntFluxtoTotFlux(1,n) = 0.0d0
      do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
        rb_RatioBandIntFluxtoTotFlux(1,n) = rb_RatioBandIntFluxtoTotFlux(1,n) &
          & + w_StrSpe(l) * DelWaveNum
      end do
      rb_RatioBandIntFluxtoTotFlux(1,n) = &
        & rb_RatioBandIntFluxtoTotFlux(1,n) / TotalFluxTOA
      do k = 1+1, NPressForPF
        rb_RatioBandIntFluxtoTotFlux(k,n) = rb_RatioBandIntFluxtoTotFlux(1,n)
      end do
    end do


    if ( FlagPFRatioOne ) then

      do l = 1, NWaveNum
        rw_AtmPFRatio(:,l) = 1.0d0
        rw_StrPFRatio(:,l) = 1.0d0
      end do

    else

      do l = 1, ab_BandWNIndexBnds(1,1)-1
        rw_AtmPFRatio(:,l) = 0.0d0
!!$        rw_StrPFRatio(:,l) = 0.0d0
      end do
      do l = ab_BandWNIndexBnds(1,1), ab_BandWNIndexBnds(2,NBand)
        DelWaveNum = w_WaveNum(2) - w_WaveNum(1)
        WNs   = w_WaveNum(l)-DelWaveNum/2.0d0
        WNe   = w_WaveNum(l)+DelWaveNum/2.0d0
        !
        GQNum = InqPFIntGQNum( DelWaveNum )
        call Integ_PF_GQ_Array1D( &
          & WNs, WNe, GQNum, &
          & 1, NPressForPF, &
          & r_TempForPF, &
          & rw_AtmPFRatio(:,l) &
          & )
        rw_AtmPFRatio(:,l) = PI * rw_AtmPFRatio(:,l)
        rw_AtmPFRatio(:,l) = rw_AtmPFRatio(:,l) / ( WNe - WNs )
        n = GetBandNum( NBand, ab_BandWNBnds, w_WaveNum(l) )
        rw_AtmPFRatio(:,l) = rw_AtmPFRatio(:,l) / rb_BandAveAtmPF(:,n)
        !
!!$        n = GetBandNum( NBand, ab_BandWNBnds, w_WaveNum(l) )
!!$        rw_StrPFRatio(:,l) = w_StrSpe(l)
!!$        rw_StrPFRatio(:,l) = rw_StrPFRatio(:,l) / rb_BandAveStrPF(:,n)

      end do
      do l = ab_BandWNIndexBnds(2,NBand)+1, NWaveNum
        rw_AtmPFRatio(:,l) = 0.0d0
!!$        rw_StrPFRatio(:,l) = 0.0d0
      end do

      ! stellar spectrum ratio in a band is assumed to be 1.
      do l = 1, NWaveNum
        rw_StrPFRatio(:,l) = 1.0d0
      end do
    end if


    ! modify absorption coefficient
    !
    if ( FlagKeyMolAbsOverlap ) then

      do l = 1, ab_BandWNIndexBnds(1,1)-1
        mw_IDAbsCoefType(:,l) = 0
      end do
      do l = ab_BandWNIndexBnds(2,NBand)+1, NWaveNum
        mw_IDAbsCoefType(:,l) = 0
      end do

      do n = 1, NBand

        mb_IDAbsCoefType(:,n) = 0
        do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
          mw_IDAbsCoefType(:,l) = 0
        end do

        allocate( pw_AbsCoefSum(NPress, NWaveNum) )
        !
        pw_AbsCoefSum = 0.0d0
        !
        do iKeyMol = 1, b_NKeyMol(n)
          do m = 1, NMol
            if ( m_MolNum(m) == ab_BandKeyMolNum(iKeyMol,n) ) exit
          end do
          if ( m > NMol ) then
            write( 6, * ) 'Unable to find key molecule, ', iKeyMol, ', in a file.'
          end if

          mb_IDAbsCoefType(m,n) = -1

          do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
            pw_AbsCoefSum(:,l) = pw_AbsCoefSum(:,l) &
              & + pmw_AbsCoef(:,m,l) * pm_VMR(:,m)
            pmw_AbsCoef(:,m,l) = 0.0d0

            mw_IDAbsCoefType(m,l) = -1
          end do

        end do

        mb_IDAbsCoefType(b_IndexFirstKeyMol(n),n) =  1
        do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
          pmw_AbsCoef(:,b_IndexFirstKeyMol(n),l) = pw_AbsCoefSum(:,l)
          mw_IDAbsCoefType(b_IndexFirstKeyMol(n),l) = 1
        end do

        deallocate( pw_AbsCoefSum )
      end do

    else
      mb_IDAbsCoefType = 0
      mw_IDAbsCoefType = 0
    end if

    allocate( wm_AbsCoef(NWaveNum, NMol) )
    allocate( w_AtmPFRatio (NWaveNum)       )
    allocate( w_StrPFRatio (NWaveNum)       )

    ! Index for a key specie
    allocate( w_SortedIndex   ( NWaveNum ) )
!!$    allocate( w_SortedIndexTmp( NWaveNum ) )
    allocate( wp_SortedIndexSave( NWaveNum, NPress ) )


    if ( FlagSort ) then

      if ( FlagCorrelation ) then
        ks = 1
        ke = NPress
      else
        ks = 1
        ke = 1
      end if

      do k = ks, ke

        ! Work variable
        do m = 1, NMol
          do l = 1, NWaveNum
            wm_AbsCoef(l,m) = pmw_AbsCoef(k,m,l)
          end do
        end do

        ! Initialization of index
        !   Index range which is sorted below will be overwritted.
        do l = 1, NWaveNum
          w_SortedIndex( l ) = l
        end do

        ! loop for band
        do n = 1, NBand

          ! sort absorption coefficient in a band
          if ( (ab_BandWNIndexBnds(2,n)-ab_BandWNIndexBnds(1,n)+1) > 0 ) then

            ! pick up a key specie in a band
            !   initialize not to miss an unset variable
!!$            IndexKeyMol = -1
!!$            do m = 1, NMol
!!$              if ( m_MolNum(m) == b_BandKeyMolNum(n) ) then
!!$                IndexKeyMol = m
!!$              end if
!!$            end do
!!$            !
!!$            m = IndexKeyMol

            m = b_IndexFirstKeyMol(n)

            call sort_heap( &
              & (ab_BandWNIndexBnds(2,n)-ab_BandWNIndexBnds(1,n)+1), &
              & wm_AbsCoef(ab_BandWNIndexBnds(1,n):ab_BandWNIndexBnds(2,n),m), &
              & w_SortedIndex(ab_BandWNIndexBnds(1,n):ab_BandWNIndexBnds(2,n)) &
              & )
            w_SortedIndex(ab_BandWNIndexBnds(1,n):ab_BandWNIndexBnds(2,n)) = &
              &   w_SortedIndex(ab_BandWNIndexBnds(1,n):ab_BandWNIndexBnds(2,n)) &
              & + ab_BandWNIndexBnds(1,n) - 1

            ! check
            do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)-1
              if ( wm_AbsCoef(l,m) > wm_AbsCoef(l+1,m) ) then
                write( 6, * ) 'Sorting check 1:'
                write( 6, * ) '  Sorting was not performed successfully.'
                write( 6, * ) '  Molecular index: ', m
                write( 6, * ) l, wm_AbsCoef(l,m), wm_AbsCoef(l+1,m)
                stop
              end if
            end do

            ! Order is reversed.
!!$            do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
!!$              w_SortedIndexTmp(l) = w_SortedIndex(l)
!!$            end do
!!$            do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
!!$              w_SortedIndex(l) = w_SortedIndexTmp(ab_BandWNIndexBnds(2,n)-(l-ab_BandWNIndexBnds(1,n)))
!!$            end do

          end if

        end do

        do l = 1, NWaveNum
          wp_SortedIndexSave(l,k) = w_SortedIndex(l)
        end do
      end do

      if ( FlagCorrelation ) then
      else
        do k = 1, NPress
          if ( k == ks ) cycle
          do l = 1, NWaveNum
            wp_SortedIndexSave(l,k) = wp_SortedIndexSave(l,ks)
          end do
        end do
      end if

    else

      do k = 1, NPress
        do l = 1, NWaveNum
          wp_SortedIndexSave(l,k) = l
        end do
      end do

    end if



    if ( FlagSort ) then

!!$      do k = 1, NPress
!!$
!!$        ! sort ac based on order of key specie
!!$        do m = 1, NMol
!!$          do l = 1, NWaveNum
!!$            wm_AbsCoef(l,m) = pmw_AbsCoef(k,m,wp_SortedIndexSave(l,k))
!!$          end do
!!$        end do
!!$        ! sort *PFRatio based on order of key specie
!!$        do l = 1, NWaveNum
!!$          w_AtmPFRatio(l) = pw_AtmPFRatio(k,wp_SortedIndexSave(l,k))
!!$          w_StrPFRatio(l) = pw_StrPFRatio(k,wp_SortedIndexSave(l,k))
!!$        end do
!!$
!!$        do m = 1, NMol
!!$          do l = 1, NWaveNum
!!$            pmw_AbsCoef(k,m,l) = wm_AbsCoef(l,m)
!!$          end do
!!$        end do
!!$        do l = 1, NWaveNum
!!$          pw_AtmPFRatio(k,l) = w_AtmPFRatio(l)
!!$          pw_StrPFRatio(k,l) = w_StrPFRatio(l)
!!$        end do
!!$
!!$      end do

      do k = 1, NPress

        ! sort ac based on order of key specie
        do m = 1, NMol
          do l = 1, NWaveNum
            wm_AbsCoef(l,m) = pmw_AbsCoef(k,m,wp_SortedIndexSave(l,k))
          end do
        end do
        do m = 1, NMol
          do l = 1, NWaveNum
            pmw_AbsCoef(k,m,l) = wm_AbsCoef(l,m)
          end do
        end do

      end do
      do k = 1, NPressForPF
        kIndex = min(k,NPress)
        if ( p_FlagCalc(kIndex) ) then
          ! sort *PFRatio based on order of key specie
          do l = 1, NWaveNum
            w_AtmPFRatio(l) = rw_AtmPFRatio(k,wp_SortedIndexSave(l,kIndex))
            ! Now, below is meaningless, because StrPFRatio is 1.
!!$            w_StrPFRatio(l) = rw_StrPFRatio(k,wp_SortedIndexSave(l,kIndex))
          end do
        else
          ! missing value is set in PFRatios, because order of key specie is invalid
          w_AtmPFRatio = PFRatioMissingValue
          ! Now, StrPFRatio is 1.
!!$          w_StrPFRatio = PFRatioMissingValue
        end if
        do l = 1, NWaveNum
          rw_AtmPFRatio(k,l) = w_AtmPFRatio(l)
!!$          rw_StrPFRatio(k,l) = w_StrPFRatio(l)
        end do
      end do


    end if

!!$    k = 1
!!$    m = 1
!!$    n = 1
!!$    do l = ab_BandWNIndexBnds(1,n), ab_BandWNIndexBnds(2,n)
!!$      write( 60, * ) l, &
!!$        & pmw_AbsCoef(k,m,l), &
!!$        & pw_AtmPFRatio(k,l), &
!!$        & pw_StrPFRatio(k,l)
!!$    end do


    deallocate( wp_SortedIndexSave )
    deallocate( w_SortedIndex    )
!!$        deallocate( w_SortedIndexTmp )


    if ( .not. FlagBinning ) then

      do n = 1, NBand
        b_BandNBin(n) = ab_BandWNIndexBnds(2,n) - ab_BandWNIndexBnds(1,n) + 1
      end do

      NBin = 0
      do n = 1, NBand
        NBin = NBin + b_BandNBin(n)
      end do

          ! this if block exists for a debug purpose
!!$          NBin = NWaveNum

      allocate( c_BinNum    (NBin) )
      allocate( ab_BandBinIndexBnds(2,NBand) )
      allocate( c_BinWeight (NBin) )
      allocate( c_BinPointWN(NBin) )
      allocate( ac_BinWNIndexBnds(2,NBin) )


      do j = 1, NBin
        c_BinNum(j) = j
      end do
      le = 0
      do n = 1, NBand
        ls = le + 1
        le = ls + b_BandNBin(n) - 1
        ab_BandBinIndexBnds(1,n) = ls
        ab_BandBinIndexBnds(2,n) = le
      end do

      c_BinPointWN(1:NBin) = w_WaveNum(ab_BandWNIndexBnds(1,1):ab_BandWNIndexBnds(2,NBand))
      c_BinWeight (1:NBin) = c_BinPointWN(2) - c_BinPointWN(1)


      ! Calculate Planck function ratio
      !
      if ( FlagSortPFRatioAverage ) then

        b_TmpBandNBin = b_ArgBandNBin
        TmpNBin = 0
        do n = 1, NBand
          TmpNBin = TmpNBin + b_TmpBandNBin(n)
        end do

        allocate( c_TmpBinNum    (NBin) )
        allocate( ab_TmpBandBinIndexBnds(2,NBand) )
        allocate( c_TmpBinWeight (NBin) )
        allocate( c_TmpBinPointWN(NBin) )
        allocate( ac_TmpBinWNIndexBnds(2,NBin) )

        call SetupBin( &
          & NWaveNum, w_WaveNum,  &
          & NBand, b_TmpBandNBin, ab_BandWNBnds, &
          & NGaussReg, a_GaussRegBnds, cb_BandGaussRegNBin, &
          & TmpNBin, &
          & c_TmpBinNum, ab_TmpBandBinIndexBnds, c_TmpBinPointWN, c_TmpBinWeight, &
          & ac_TmpBinWNIndexBnds &
          & )

        do k = 1, NPressForPF
          do j = 1, TmpNBin

            ls = ac_TmpBinWNIndexBnds(1,j)
            le = ac_TmpBinWNIndexBnds(2,j)
            !
            Sum = 0.0d0
            do l = ls, le
              Sum = Sum + rw_AtmPFRatio(k,l)
            end do
            if ( ls <= le ) then
              Sum = Sum / ( le - ls + 1 )
            else
              Sum = 1.0d0
            end if
            do l = ls, le
              rw_AtmPFRatio(k,l) = Sum
            end do
            !
            ! Now, StrPFRatio is 1.
            Sum = 0.0d0
            do l = ls, le
              Sum = Sum + rw_StrPFRatio(k,l)
            end do
            if ( ls <= le ) then
              Sum = Sum / ( le - ls + 1 )
            else
              Sum = 1.0d0
            end if
            do l = ls, le
              rw_StrPFRatio(k,l) = Sum
            end do

          end do
        end do

        deallocate( c_TmpBinNum            )
        deallocate( ab_TmpBandBinIndexBnds )
        deallocate( c_TmpBinWeight         )
        deallocate( c_TmpBinPointWN        )
        deallocate( ac_TmpBinWNIndexBnds   )



!!$        allocate( c_BinWNBnds(0:NBin) )
!!$
!!$        ! Actually, bin is not bounded by wavenumber, but is bounded by 
!!$        ! cumulative probability.
!!$        ! But, here, wavenumber bound is considered for clarity.
!!$        !
!!$        c_BinWNBnds(0) = ab_BandWNBnds(1,1)
!!$        do j = 1, NBin
!!$          c_BinWNBnds(j) = c_BinWNBnds(j-1) + c_BinWeight(j)
!!$        end do
!!$
!!$        ! Look for bin bound index
!!$        search_bin_bound_index_1 : do l = 1+1, NWaveNum
!!$          if ( w_WaveNum(l) >= c_BinWNBnds(0) ) exit search_bin_bound_index_1
!!$        end do search_bin_bound_index_1
!!$        ac_BinWNIndexBnds(1,1) = l
!!$        !
!!$        do j = 1, NBin
!!$          search_bin_bound_index : do l = ac_BinWNIndexBnds(1,j), NWaveNum
!!$            if ( w_WaveNum(l) > c_BinWNBnds(j) ) exit search_bin_bound_index
!!$          end do search_bin_bound_index
!!$          ac_BinWNIndexBnds(2,j) = l-1
!!$          if ( j < NBin ) then
!!$            ac_BinWNIndexBnds(1,j+1) = ac_BinWNIndexBnds(2,j) + 1
!!$          end if
!!$        end do
!!$
!!$
!!$        do k = 1, NPress
!!$          do n = 1, NBand
!!$
!!$            do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
!!$              ls = ac_BinWNIndexBnds(1,j)
!!$              le = ac_BinWNIndexBnds(2,j)
!!$              !
!!$              Sum = 0.0d0
!!$              do l = ls, le
!!$                Sum = Sum + pw_AtmPFRatio(k,l)
!!$              end do
!!$              if ( ls <= le ) then
!!$                Sum = Sum / ( le - ls + 1 )
!!$              else
!!$                Sum = 1.0d0
!!$              end if
!!$              do l = ls, le
!!$                pw_AtmPFRatio(k,l) = Sum
!!$              end do
!!$              !
!!$              Sum = 0.0d0
!!$              do l = ls, le
!!$                Sum = Sum + pw_StrPFRatio(k,l)
!!$              end do
!!$              if ( ls <= le ) then
!!$                Sum = Sum / ( le - ls + 1 )
!!$              else
!!$                Sum = 1.0d0
!!$              end if
!!$              do l = ls, le
!!$                pw_StrPFRatio(k,l) = Sum
!!$              end do
!!$            end do
!!$
!!$          end do
!!$        end do
!!$
!!$        deallocate( c_BinWNBnds )

      end if


      allocate( pmc_BinAbsCoef(NPress, NMol, NBin) )
      allocate( rc_BinAtmPFRatio (NPressForPF, NBin) )
      allocate( rc_BinStrPFRatio (NPressForPF, NBin) )

      pmc_BinAbsCoef  (:,:,1:NBin) = pmw_AbsCoef  (:,:,ab_BandWNIndexBnds(1,1):ab_BandWNIndexBnds(2,NBand))
      rc_BinAtmPFRatio(:,1:NBin)   = rw_AtmPFRatio(:,ab_BandWNIndexBnds(1,1):ab_BandWNIndexBnds(2,NBand))
      rc_BinStrPFRatio(:,1:NBin)   = rw_StrPFRatio(:,ab_BandWNIndexBnds(1,1):ab_BandWNIndexBnds(2,NBand))

      allocate( mc_IDAbsCoefType(NMol, NBin) )

      do n = 1, NBand
        do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
          mc_IDAbsCoefType(:,j) = mb_IDAbsCoefType(:,n)
        end do
      end do

    else

      b_BandNBin = b_ArgBandNBin

      NBin = 0
      do n = 1, NBand
        NBin = NBin + b_BandNBin(n)
      end do

      allocate( c_BinNum    (NBin) )
      allocate( ab_BandBinIndexBnds(2,NBand) )
      allocate( c_BinWeight (NBin) )
      allocate( c_BinPointWN(NBin) )
      allocate( ac_BinWNIndexBnds(2,NBin) )

      call SetupBin( &
        & NWaveNum, w_WaveNum,  &
        & NBand, b_BandNBin, ab_BandWNBnds, &
        & NGaussReg, a_GaussRegBnds, cb_BandGaussRegNBin, &
        & NBin, &
        & c_BinNum, ab_BandBinIndexBnds, c_BinPointWN, c_BinWeight, &
        & ac_BinWNIndexBnds &
        & )

      allocate( pmc_BinAbsCoef(NPress, NMol, NBin) )
      allocate( rc_BinAtmPFRatio (NPressForPF, NBin) )
      allocate( rc_BinStrPFRatio (NPressForPF, NBin) )


      do k = 1, NPress
        do n = 1, NBand

          ! Calculate absorption coefficient of a key molecule
          !
          !   pick up a key specie in a band
          !
          !     initialize not to miss an unset variable
!!$          IndexKeyMol = -1
!!$          do m = 1, NMol
!!$            if ( m_MolNum(m) == b_BandKeyMolNum(n) ) then
!!$              IndexKeyMol = m
!!$            end if
!!$          end do
          !
          if ( FlagKeyMolACAverage ) then

!!$            m = IndexKeyMol
            m = b_IndexFirstKeyMol(n)
            !
            do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
              ls = ac_BinWNIndexBnds(1,j)
              le = ac_BinWNIndexBnds(2,j)
              if ( ls <= le ) then
                Sum = 0.0d0
                do l = ls, le
                  Sum = Sum + pmw_AbsCoef(k,m,l)
                end do
                Sum = Sum / ( le - ls + 1 )
              else
                Sum = 0.0d0
              end if
              pmc_BinAbsCoef(k,m,j) = Sum
            end do

          else

            if ( .not. FlagSort ) then
              write( 6, * ) 'This is not supported, tag#110.'
              write( 6, * ) '  If FlagBinning is true, following combinations are allowed.'
              write( 6, * ) '    FlagBinning = .true. & FlagSort is .true.'
              write( 6, * ) '    FlagBinning = .true. & FlagKeyMolACAverage is .true.'
              stop
            end if


!!$            m = IndexKeyMol
            m = b_IndexFirstKeyMol(n)

            if ( (ab_BandBinIndexBnds(2,n)-ab_BandBinIndexBnds(1,n)+1) > 0 ) then

              ls = ab_BandWNIndexBnds(1,n)+1
              do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
                if ( c_BinPointWN(j) < w_WaveNum(ab_BandWNIndexBnds(1,n)) ) then
                  l = ab_BandWNIndexBnds(1,n)
                  pmc_BinAbsCoef(k,m,j) = pmw_AbsCoef(k,m,l)
                else if ( w_WaveNum(ab_BandWNIndexBnds(2,n)) < c_BinPointWN(j) ) then
                  l = ab_BandWNIndexBnds(2,n)
                  pmc_BinAbsCoef(k,m,j) = pmw_AbsCoef(k,m,l)
                else
!!$                  search_gausspoint : do l = ls, NWaveNum-1
                  search_gausspoint : do l = ls, ab_BandWNIndexBnds(2,n)-1
                    if ( c_BinPointWN(j) < w_WaveNum(l) ) then
                      exit search_gausspoint
                    end if
                  end do search_gausspoint
                  pmc_BinAbsCoef(k,m,j) = &
                    &   ( pmw_AbsCoef(k,m,l) - pmw_AbsCoef(k,m,l-1) ) &
                    & / ( w_WaveNum(l)       - w_WaveNum(l-1)       ) &
                    & * ( c_BinPointWN(j)    - w_WaveNum(l-1)       ) &
                    & + pmw_AbsCoef(k,m,l-1)
                  pmc_BinAbsCoef(k,m,j) = max( pmc_BinAbsCoef(k,m,j), 0.0d0 )
                  ls = l
                end if
              end do
            end if

          end if

          ! Calculate absorption coefficient of non-key molecules by average
          !
          do m = 1, NMol
!!$            if ( m == IndexKeyMol ) cycle
            if ( m == b_IndexFirstKeyMol(n) ) cycle

            if ( FlagNonKeyMolACAverage ) then

              do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
                ! setting for bin averaged abs. coef. for non-key molecules
                ls = ac_BinWNIndexBnds(1,j)
                le = ac_BinWNIndexBnds(2,j)
!!$                ! setting for band averaged abs. coef. for non-key molecules
!!$                ls = ab_BandWNIndexBnds(1,n)
!!$                le = ab_BandWNIndexBnds(2,n)
                if ( ls <= le ) then
                  Sum = 0.0d0
                  do l = ls, le
                    Sum = Sum + pmw_AbsCoef(k,m,l)
                  end do
                  Sum = Sum / ( le - ls + 1 )
                else
                  Sum = 0.0d0
                end if
                pmc_BinAbsCoef(k,m,j) = Sum
              end do

            else

              if ( .not. FlagSort ) then
                write( 6, * ) 'This is not supported, tag#120.'
                write( 6, * ) '  If FlagBinning is true, following combinations are allowed.'
                write( 6, * ) '    FlagBinning = .true. & FlagSort is .true.'
                write( 6, * ) '    FlagBinning = .true. & FlagNonKeyMolACAverage is .true.'
                stop
              end if

              if ( (ab_BandBinIndexBnds(2,n)-ab_BandBinIndexBnds(1,n)+1) > 0 ) then

                ls = ab_BandWNIndexBnds(1,n)+1
                do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
                  if ( c_BinPointWN(j) < w_WaveNum(ab_BandWNIndexBnds(1,n)) ) then
                    l = ab_BandWNIndexBnds(1,n)
                    pmc_BinAbsCoef(k,m,j) = pmw_AbsCoef(k,m,l)
                  else if ( w_WaveNum(ab_BandWNIndexBnds(2,n)) < c_BinPointWN(j) ) then
                    l = ab_BandWNIndexBnds(2,n)
                    pmc_BinAbsCoef(k,m,j) = pmw_AbsCoef(k,m,l)
                  else
!!$                    search_gausspoint_nonkey : do l = ls, NWaveNum-1
                    search_gausspoint_nonkey : do l = ls, ab_BandWNIndexBnds(2,n)-1
                      if ( c_BinPointWN(j) < w_WaveNum(l) ) then
                        exit search_gausspoint_nonkey
                      end if
                    end do search_gausspoint_nonkey
                    pmc_BinAbsCoef(k,m,j) = &
                      &   ( pmw_AbsCoef(k,m,l) - pmw_AbsCoef(k,m,l-1) ) &
                      & / ( w_WaveNum(l)       - w_WaveNum(l-1)       ) &
                      & * ( c_BinPointWN(j)    - w_WaveNum(l-1)       ) &
                      & + pmw_AbsCoef(k,m,l-1)
                    pmc_BinAbsCoef(k,m,j) = max( pmc_BinAbsCoef(k,m,j), 0.0d0 )
                    ls = l
                  end if
                end do
              end if

            end if

          end do

        end do
      end do

      do k = 1, NPressForPF
        do n = 1, NBand

          ! Calculate Planck function ratio
          !
          if ( FlagSortPFRatioAverage ) then

            do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
              ls = ac_BinWNIndexBnds(1,j)
              le = ac_BinWNIndexBnds(2,j)
              !
              if ( ls <= le ) then
                Sum = 0.0d0
                do l = ls, le
                  Sum = Sum + rw_AtmPFRatio(k,l)
                end do
                Sum = Sum / ( le - ls + 1 )
              else
                Sum = 1.0d0
              end if
              rc_BinAtmPFRatio(k,j) = Sum
              !
              ! Now, StrPFRatio is 1.
              if ( ls <= le ) then
                Sum = 0.0d0
                do l = ls, le
                  Sum = Sum + rw_StrPFRatio(k,l)
                end do
                Sum = Sum / ( le - ls + 1 )
              else
                Sum = 1.0d0
              end if
              rc_BinStrPFRatio(k,j) = Sum
            end do

          else

            write( 6, * ) 'In ', trim( ModuleName )
            write( 6, * ) 'This is not supported, now.'
            stop

!!$            ! sort absorption coefficient in a bin
!!$            if ( (ab_BandBinIndexBnds(2,n)-ab_BandBinIndexBnds(1,n)+1) > 0 ) then
!!$              call sort_heap( &
!!$                & (ab_BandBinIndexBnds(2,n)-ab_BandBinIndexBnds(1,n)+1), &
!!$                & w_AtmPFRatio(ab_BandBinIndexBnds(1,n):ab_BandBinIndexBnds(2,n)) &
!!$                & )
!!$              call sort_heap( &
!!$                & (ab_BandBinIndexBnds(2,n)-ab_BandBinIndexBnds(1,n)+1), &
!!$                & w_StrPFRatio(ab_BandBinIndexBnds(1,n):ab_BandBinIndexBnds(2,n)) &
!!$                & )
!!$              !
!!$              ls = 2
!!$              do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
!!$                search_gausspoint_pf : do l = ls, NWaveNum-1
!!$                  if ( c_BinPointWN(j) < w_WaveNum(l) ) then
!!$                    exit search_gausspoint_pf
!!$                  end if
!!$                end do search_gausspoint_pf
!!$                pc_BinAtmPFRatio(k,j) = &
!!$                  &   ( pw_AtmPFRatio(k,l) - pw_AtmPFRatio(k,l-1) ) &
!!$                  & / ( w_WaveNum(l)       - w_WaveNum(l-1)       ) &
!!$                  & * ( c_BinPointWN(j)    - w_WaveNum(l-1)       ) &
!!$                  & + pw_AtmPFRatio(k,l-1)
!!$                pc_BinAtmPFRatio(k,j) = max( pc_BinAtmPFRatio(k,j), 0.0d0 )
!!$                !
!!$                pc_BinStrPFRatio(k,j) = &
!!$                  &   ( pw_StrPFRatio(k,l) - pw_StrPFRatio(k,l-1) ) &
!!$                  & / ( w_WaveNum(l)       - w_WaveNum(l-1)       ) &
!!$                  & * ( c_BinPointWN(j)    - w_WaveNum(l-1)       ) &
!!$                  & + pw_StrPFRatio(k,l-1)
!!$                pc_BinStrPFRatio(k,j) = max( pc_BinStrPFRatio(k,j), 0.0d0 )
!!$                ls = l
!!$              end do
!!$            end if

          end if


        end do
      end do


      allocate( mc_IDAbsCoefType(NMol, NBin) )

      do n = 1, NBand
        do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
          mc_IDAbsCoefType(:,j) = mb_IDAbsCoefType(:,n)
        end do
      end do


      if ( FlagBinningDemonstration ) then

        ! Mapping bin again to retain accuracy

        allocate( c_TmpBinNum    (NBin) )
        allocate( c_TmpBinWeight (NBin) )
        allocate( c_TmpBinPointWN(NBin) )
        allocate( pmc_TmpBinAbsCoef  (NPress, NMol, NBin) )
        allocate( rc_TmpBinAtmPFRatio   (NPressForPF, NBin) )
        allocate( rc_TmpBinStrPFRatio   (NPressForPF, NBin) )
        allocate( mc_TmpIDAbsCoefType(NMol, NBin) )
        !
        c_TmpBinNum         = c_BinNum
        c_TmpBinWeight      = c_BinWeight
        c_TmpBinPointWN     = c_BinPointWN
        pmc_TmpBinAbsCoef   = pmc_BinAbsCoef
        rc_TmpBinAtmPFRatio = rc_BinAtmPFRatio
        rc_TmpBinStrPFRatio = rc_BinStrPFRatio
        mc_TmpIDAbsCoefType = mc_IDAbsCoefType
        !
        deallocate( c_BinNum         )
        deallocate( c_BinWeight      )
        deallocate( c_BinPointWN     )
        deallocate( pmc_BinAbsCoef   )
        deallocate( rc_BinAtmPFRatio )
        deallocate( rc_BinStrPFRatio )
        deallocate( mc_IDAbsCoefType )
        !
        do n = 1, NBand
          b_BandNBin(n) = ab_BandWNIndexBnds(2,n) - ab_BandWNIndexBnds(1,n) + 1
        end do
        NBin = 0
        do n = 1, NBand
          NBin = NBin + b_BandNBin(n)
        end do
        !
        allocate( c_BinNum    (NBin) )
        allocate( c_BinWeight (NBin) )
        allocate( c_BinPointWN(NBin) )
        allocate( pmc_BinAbsCoef(NPress, NMol, NBin) )
        allocate( rc_BinAtmPFRatio (NPressForPF, NBin) )
        allocate( rc_BinStrPFRatio (NPressForPF, NBin) )
        allocate( mc_IDAbsCoefType(NMol, NBin) )
        !
        c_BinPointWN(1:NBin) = w_WaveNum(ab_BandWNIndexBnds(1,1):ab_BandWNIndexBnds(2,NBand))
!!$            c_BinWeight (1:NBin) = c_BinPointWN(2) - c_BinPointWN(1)
        do n = 1, NBand
          do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
            ls = ac_BinWNIndexBnds(1,j) - ac_BinWNIndexBnds(1,1) + 1
            le = ac_BinWNIndexBnds(2,j) - ac_BinWNIndexBnds(1,1) + 1
            do l = ls, le
              c_BinWeight(l) = c_TmpBinWeight(j) / ( le - ls + 1 )
            end do
          end do
        end do
        do n = 1, NBand
          do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
            ls = ac_BinWNIndexBnds(1,j) - ac_BinWNIndexBnds(1,1) + 1
            le = ac_BinWNIndexBnds(2,j) - ac_BinWNIndexBnds(1,1) + 1
            do l = ls, le
              do m = 1, NMol
                do k = 1, NPress
                  pmc_BinAbsCoef(k, m, l) = pmc_TmpBinAbsCoef(k, m, j)
                end do
              end do
              do k = 1, NPressForPF
                rc_BinAtmPFRatio  (k, l) = rc_TmpBinAtmPFRatio  (k, j)
                rc_BinStrPFRatio  (k, l) = rc_TmpBinStrPFRatio  (k, j)
              end do
            end do
          end do
        end do
        !
        do n = 1, NBand
          do j = ab_BandBinIndexBnds(1,n), ab_BandBinIndexBnds(2,n)
            ls = ac_BinWNIndexBnds(1,j) - ac_BinWNIndexBnds(1,1) + 1
            le = ac_BinWNIndexBnds(2,j) - ac_BinWNIndexBnds(1,1) + 1
            do l = ls, le
              mc_IDAbsCoefType(:,l) = mb_IDAbsCoefType(:,n)
            end do
          end do
        end do
        !
        ! NOTE:
        !   ab_BandBinIndexBnds is overwritten below, since it is used above.
        !
        do j = 1, NBin
          c_BinNum(j) = j
        end do
        le = 0
        do n = 1, NBand
          ls = le + 1
          le = ls + b_BandNBin(n) - 1
          ab_BandBinIndexBnds(1,n) = ls
          ab_BandBinIndexBnds(2,n) = le
        end do
        !
        deallocate( c_TmpBinNum         )
        deallocate( c_TmpBinWeight      )
        deallocate( c_TmpBinPointWN     )
        deallocate( pmc_TmpBinAbsCoef   )
        deallocate( rc_TmpBinAtmPFRatio    )
        deallocate( rc_TmpBinStrPFRatio    )
        deallocate( mc_TmpIDAbsCoefType )

      end if

    end if

    deallocate( wm_AbsCoef )
    deallocate( w_AtmPFRatio  )
    deallocate( w_StrPFRatio  )



    ! overwrite absorption coefficient where missing value is set
    do k = 1, NPress
      if ( .not. p_FlagCalc(k) ) then
        do l = 1, NWaveNum
          do m = 1, NMol
            pmw_AbsCoef(k,m,l) = AbsCoefMissingValue
          end do
        end do
        do j = 1, NBin
          do m = 1, NMol
            pmc_BinAbsCoef(k,m,j) = AbsCoefMissingValue
          end do
        end do
      end if
    end do
    ! overwrite absorption coefficient where missing value is set
    do k = 1, NPressForPF
      kIndex = min(k,NPress)
      if ( .not. p_FlagCalc(kIndex) ) then
        do l = 1, NWaveNum
          rw_AtmPFRatio(kIndex,l) = PFRatioMissingValue
          ! Now, StrPFRatio is 1.
!!$          rw_StrPFRatio(kIndex,l) = PFRatioMissingValue
        end do
        do j = 1, NBin
          rc_BinAtmPFRatio(kIndex,j) = PFRatioMissingValue
          ! Now, StrPFRatio is 1.
!!$          rc_BinStrPFRatio(kIndex,j) = PFRatioMissingValue
        end do
      end if
    end do


    ! Output

    Mode = "new"
    call ni3_open( OutNcFN, Mode, NcID )

    call ni3_put_att( NcID, '_global_', 'comment_dwn_m-1', CommentDWN )
    call ni3_put_att( NcID, '_global_', 'comment_strspe', CommentStrSpe )
!!$    call ni3_put_att( NcID, '_global_', 'pressure_for_P.F._sort_Pa', p_Press(kIndexForPFSort) )

    Name     = "Press"
    StdName  = "plev"
    LongName = "pressure"
    Units    = "Pa"
    call ni3_set_dim( NcID, Name, NF90_DOUBLE, p_Press  , &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "MolNum"
    StdName  = "MolNum"
    LongName = "molecular number"
    Units    = "1"
    call ni3_set_dim( NcID, Name, NF90_INT   , m_MolNum, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    if ( .not. FlagBinning ) then
      Name     = "WaveNum"
      StdName  = "wavenumber"
      LongName = "wavenumber"
      Units    = "m-1"
!!$      call ni3_set_dim( NcID, Name, NF90_DOUBLE, w_WaveNum, &
!!$        & stdname = StdName, longname = LongName, units  = Units )
      call ni3_def_dim( NcID, Name, NF90_DOUBLE, NI3_UNLIMITED, &
        & stdname = StdName, longname = LongName, units  = Units )
      do l = 1, NWaveNum
        call ni3_put_varss( NcID, Name, l, w_WaveNum(l) )
      end do
    end if
    !
    Name     = "BandNum"
    StdName  = "BandNum"
    LongName = "band number"
    Units    = "1"
    call ni3_set_dim( NcID, Name, NF90_INT   , b_BandNum, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "BandWaveNum"
    StdName  = "BandWaveNum"
    LongName = "band wavenumber"
    Units    = "m-1"
    call ni3_set_dim( NcID, Name, NF90_DOUBLE, b_BandWN, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "BinNum"
    StdName  = "BinNum"
    LongName = "bin number"
    Units    = "1"
    call ni3_set_dim( NcID, Name, NF90_INT   , c_BinNum, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "BinWaveNum"
    StdName  = "BinWaveNum"
    LongName = "bin wavenumber"
    Units    = "m-1"
    call ni3_set_dim( NcID, Name, NF90_DOUBLE, c_BinPointWN, &
      & stdname = StdName, longname = LongName, units  = Units )
    Comment = "This dimension is present for visualization purpose. This can be replaced with BinNum dimension."
    call ni3_put_att( NcID, Name, "comment", Comment )
    !
    Name     = "BndNum"
    StdName  = "BndNum"
    LongName = "bound number"
    Units    = "1"
    call ni3_set_dim( NcID, Name, NF90_INT   , a_BndNum, &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    Name     = "PressForPF"
    StdName  = "plev"
    LongName = "pressure"
    Units    = "Pa"
    call ni3_set_dim( NcID, Name, NF90_DOUBLE, r_PressForPF  , &
      & stdname = StdName, longname = LongName, units  = Units )
    !
    !
    NDims = 1
    allocate( a_DimNames( NDims ) )
    !
    a_DimNames( 1 ) = "Press"
    !
    Name     = "Temp"
    StdName  = Name
    LongName = "air_temperature"
    Units    = "K"
    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, p_Temp )
    !
    deallocate( a_DimNames )
    !
    !
    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "Press"
    a_DimNames( 2 ) = "MolNum"
    Name     = "VMR"
    StdName  = Name
    LongName = "volume mixing ratio"
    Units    = "1"
    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, pm_VMR )
    deallocate( a_DimNames )
    !
    if ( .not. FlagBinning ) then
      NDims = 3
      allocate( a_DimNames( NDims ) )
      a_DimNames( 1 ) = "Press"
      a_DimNames( 2 ) = "MolNum"
      a_DimNames( 3 ) = "WaveNum"
      Name     = AbsCoefVarName
      StdName  = Name
      LongName = "absorption_coefficient"
      Units    = "(molecule m-2)-1"
      if ( FlagMissingValue ) then
        call ni3_def_var( NcID, name, NF90_DOUBLE, NDims, a_DimNames, &
          & stdname = StdName, longname = LongName, units = Units, fillvalue = AbsCoefMissingValue )
        call ni3_put_att( NcID, Name, "missing_value", AbsCoefMissingValue )
      else
        call ni3_def_var( NcID, name, NF90_DOUBLE, NDims, a_DimNames, &
          & stdname = StdName, longname = LongName, units = Units )
      end if
      call ni3_put_var( NcID, Name, pmw_AbsCoef )
      deallocate( a_DimNames )
      !
      NDims = 2
      allocate( a_DimNames( NDims ) )
      a_DimNames( 1 ) = "PressForPF"
      a_DimNames( 2 ) = "WaveNum"
      Name     = "AtmPFRatio"
      StdName  = Name
      LongName = "Planck function ratio"
      Units    = "1"
      if ( FlagMissingValue ) then
        call ni3_def_var( NcID, name, NF90_DOUBLE, NDims, a_DimNames, &
          & stdname = StdName, longname = LongName, units = Units, fillvalue = PFRatioMissingValue )
        call ni3_put_att( NcID, Name, "missing_value", PFRatioMissingValue )
      else
        call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
          & stdname = StdName, longname = LongName, units = Units )
      end if
      call ni3_put_var( NcID, Name, rw_AtmPFRatio )
      deallocate( a_DimNames )
      !
!!$      NDims = 2
!!$      allocate( a_DimNames( NDims ) )
!!$      a_DimNames( 1 ) = "PressForPF"
!!$      a_DimNames( 2 ) = "WaveNum"
!!$      Name     = "StrPFRatio"
!!$      StdName  = Name
!!$      LongName = "Planck function ratio"
!!$      Units    = "1"
!!$      if ( FlagMissingValue ) then
!!$        call ni3_def_var( NcID, name, NF90_DOUBLE, NDims, a_DimNames, &
!!$          & stdname = StdName, longname = LongName, units = Units, fillvalue = PFRatioMissingValue )
!!$        call ni3_put_att( NcID, Name, "missing_value", PFRatioMissingValue )
!!$      else
!!$        call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
!!$          & stdname = StdName, longname = LongName, units = Units )
!!$      end if
!!$      call ni3_put_var( NcID, Name, rw_StrPFRatio )
!!$      deallocate( a_DimNames )
      !
      ! Now, StrPFRatio is 1.
      allocate  ( w_StrPFRatio(NWaveNum) )
      do l = 1, NWaveNum
        w_StrPFRatio(l) = rw_StrPFRatio(1,l)
      end do
      NDims = 1
      allocate( a_DimNames( NDims ) )
      a_DimNames( 1 ) = "WaveNum"
      Name     = "StrPFRatio"
      StdName  = Name
      LongName = "Planck function ratio"
      Units    = "1"
      call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
      call ni3_put_var( NcID, Name, w_StrPFRatio )
      deallocate( a_DimNames )
      deallocate( w_StrPFRatio           )
    end if



    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BndNum"
    a_DimNames( 2 ) = "BandNum"
    Name     = "BandWaveNumBnds"
    StdName  = Name
    LongName = "band wavenumber bound"
    Units    = "m-1"
    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, ab_BandWNBnds )
    deallocate( a_DimNames )


!!$    if ( .not. FlagBinning ) then
!!$      NDims = 2
!!$      allocate( a_DimNames( NDims ) )
!!$      a_DimNames( 1 ) = "BndNum"
!!$      a_DimNames( 2 ) = "BandNum"
!!$      Name     = "BandWaveNumIndexBnds"
!!$      StdName  = Name
!!$      LongName = "band index bound"
!!$      Units    = "1"
!!$      call ni3_def_var( NcID, Name, NF90_INT, NDims, a_DimNames, &
!!$        & stdname = StdName, longname = LongName, units = Units )
!!$      call ni3_put_var( NcID, Name, ab_BandWNIndexBnds )
!!$      deallocate( a_DimNames )
!!$    end if

    NDims = 1
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BandNum"
    Name     = "BandNBin"
    StdName  = Name
    LongName = "number of bins in a band"
    Units    = "1"
    call ni3_def_var( NcID, Name, NF90_INT, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, b_BandNBin )
    deallocate( a_DimNames )


    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BndNum"
    a_DimNames( 2 ) = "BandNum"
    Name     = "BandBinIndexBnds"
    StdName  = Name
    LongName = "index bound of bins in a band"
    Units    = "1"
    call ni3_def_var( NcID, Name, NF90_INT, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, ab_BandBinIndexBnds )
    deallocate( a_DimNames )


!!$    NDims = 1
!!$    allocate( a_DimNames( NDims ) )
!!$    a_DimNames( 1 ) = "BandNum"
!!$    Name     = "BandKeyMolNum"
!!$    StdName  = Name
!!$    LongName = "band key molecular number"
!!$    Units    = "1"
!!$    call ni3_def_var( NcID, Name, NF90_INT, NDims, a_DimNames, &
!!$      & stdname = StdName, longname = LongName, units = Units )
!!$    call ni3_put_var( NcID, Name, b_BandKeyMolNum )
!!$    deallocate( a_DimNames )

    NDims = 1
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BinNum"
    Name     = "BinWeight"
    StdName  = Name
    LongName = "bin weight"
    Units    = "m-1"
    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, c_BinWeight )
    deallocate( a_DimNames )


    NDims = 3
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "Press"
    a_DimNames( 2 ) = "MolNum"
    a_DimNames( 3 ) = "BinWaveNum"
    Name     = "BinAbsCoef"
    StdName  = Name
    LongName = "absorption_coefficient"
    Units    = "(molecule m-2)-1"
    if ( FlagMissingValue ) then
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units, fillvalue = AbsCoefMissingValue )
      call ni3_put_att( NcID, Name, "missing_value", AbsCoefMissingValue )
    else
      call ni3_def_var( NcID, name, NF90_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
    end if
    call ni3_put_var( NcID, Name, pmc_BinAbsCoef )
    deallocate( a_DimNames )
    !
    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "PressForPF"
    a_DimNames( 2 ) = "BinWaveNum"
    Name     = "BinAtmPFRatio"
    StdName  = Name
    LongName = "Planck function ratio"
    Units    = "1"
    if ( FlagMissingValue ) then
      call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units, fillvalue = PFRatioMissingValue )
      call ni3_put_att( NcID, Name, "missing_value", PFRatioMissingValue )
    else
      call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
    end if
    call ni3_put_var( NcID, Name, rc_BinAtmPFRatio )
    deallocate( a_DimNames )
    !
!!$    NDims = 2
!!$    allocate( a_DimNames( NDims ) )
!!$    a_DimNames( 1 ) = "Press"
!!$    a_DimNames( 2 ) = "BandNum"
!!$    Name     = "BandAveStrPF"
!!$    StdName  = Name
!!$    LongName = "Band averaged star flux"
!!$    Units    = "1"
!!$    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
!!$      & stdname = StdName, longname = LongName, units = Units )
!!$    call ni3_put_var( NcID, Name, pb_BandAveStrPF )
!!$    deallocate( a_DimNames )
    !
    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "PressForPF"
!!$    a_DimNames( 2 ) = "BandNum"
    a_DimNames( 2 ) = "BandWaveNum"
    Name     = "RatioBandIntFluxtoTotFlux"
    StdName  = Name
    LongName = "Ratio of band integrated flux to total flux"
    Units    = "1"
    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, rb_RatioBandIntFluxtoTotFlux )
    deallocate( a_DimNames )
    !
!!$    NDims = 2
!!$    allocate( a_DimNames( NDims ) )
!!$    a_DimNames( 1 ) = "PressForPF"
!!$    a_DimNames( 2 ) = "BinWaveNum"
!!$    Name     = "BinStrPFRatio"
!!$    StdName  = Name
!!$    LongName = "Planck function ratio"
!!$    Units    = "1"
!!$    if ( FlagMissingValue ) then
!!$      call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
!!$        & stdname = StdName, longname = LongName, units = Units, fillvalue = PFRatioMissingValue )
!!$      call ni3_put_att( NcID, Name, "missing_value", PFRatioMissingValue )
!!$    else
!!$      call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
!!$        & stdname = StdName, longname = LongName, units = Units )
!!$    end if
!!$    call ni3_put_var( NcID, Name, rc_BinStrPFRatio )
!!$    deallocate( a_DimNames )
    !
    ! Now, StrPFRatio is 1.
    allocate  ( c_BinStrPFRatio(NBin) )
    do j = 1, NBin
      c_BinStrPFRatio(j) = rc_BinStrPFRatio(1,j)
    end do
    NDims = 1
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "BinWaveNum"
    Name     = "BinStrPFRatio"
    StdName  = Name
    LongName = "Planck function ratio"
    Units    = "1"
    call ni3_def_var( NcID, Name, NF90_DOUBLE, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    call ni3_put_var( NcID, Name, c_BinStrPFRatio )
    deallocate( a_DimNames )
    deallocate( c_BinStrPFRatio )

    if ( .not. FlagBinning ) then
      NDims = 2
      allocate( a_DimNames( NDims ) )
      a_DimNames( 1 ) = "MolNum"
      a_DimNames( 2 ) = "WaveNum"
      Name     = "IDAbsCoefType"
      StdName  = Name
      LongName = "absorption type ID"
      Units    = "1"
      call ni3_def_var( NcID, Name, NF90_INT, NDims, a_DimNames, &
        & stdname = StdName, longname = LongName, units = Units )
      Comment = "-1: zero abs. coef., " &
        &    // "0: abs. coef. per unit abs. molecule, " &
        &    // "1: abs. coef. per unit atm. molecule"
      call ni3_put_att( NcID, Name, "comment", Comment )
      call ni3_put_var( NcID, Name, mw_IDAbsCoefType )
      deallocate( a_DimNames )
    end if

    NDims = 2
    allocate( a_DimNames( NDims ) )
    a_DimNames( 1 ) = "MolNum"
    a_DimNames( 2 ) = "BinNum"
    Name     = "BinIDAbsCoefType"
    StdName  = Name
    LongName = "absorption type ID"
    Units    = "1"
    call ni3_def_var( NcID, Name, NF90_INT, NDims, a_DimNames, &
      & stdname = StdName, longname = LongName, units = Units )
    Comment = "-1: zero abs. coef., " &
      &    // "0: abs. coef. per unit abs. molecule, " &
      &    // "1: abs. coef. per unit atm. molecule"
    call ni3_put_att( NcID, Name, "comment", Comment )
    call ni3_put_var( NcID, Name, mc_IDAbsCoefType )
    deallocate( a_DimNames )

    call ni3_close( NcID )



    deallocate( rb_RatioBandIntFluxtoTotFlux )
    deallocate( rb_BandAveAtmPF   )
    deallocate( rb_BandAveStrPF   )
    deallocate( rc_BinAtmPFRatio  )
    deallocate( rc_BinStrPFRatio  )
    deallocate( pmc_BinAbsCoef    )
    deallocate( ac_BinWNIndexBnds )
    deallocate( c_BinPointWN      )
    deallocate( c_BinWeight       )
    deallocate( ab_BandBinIndexBnds )
    deallocate( c_BinNum          )

    deallocate( b_BandWN )
    deallocate( w_StrSpe )
    deallocate( pm_VMR )
    deallocate( p_FlagCalc  )
    deallocate( pmw_AbsCoef )
    deallocate( rw_AtmPFRatio )
    deallocate( rw_StrPFRatio )

    deallocate( w_WaveNum )
    deallocate( m_MolNum  )
    deallocate( p_Press   )
    deallocate( p_Temp    )

    deallocate( mb_IDAbsCoefType )
    deallocate( mw_IDAbsCoefType )
    deallocate( mc_IDAbsCoefType )

    deallocate( ab_BandWNIndexBnds )
    deallocate( b_BandNum )

  end subroutine SortAC

  !--------------------------------------------------------------------------

  function GetBandNum( NBand, ab_WNBnds, WaveNum ) result( BandNum )

    integer , intent(in ) :: NBand
    real(DP), intent(in ) :: ab_WNBnds(2,NBand)
    real(DP), intent(in ) :: WaveNum

    integer :: BandNum

    integer :: n


!!$    if ( .not. FlagInited ) then
!!$      write( 6, * ) trim( ModuleName ), " is not initialized."
!!$      stop
!!$    end if

    search_Band : do n = 1, NBand
       if ( WaveNum <= ab_WNBnds(2,n) ) exit search_Band
    end do search_Band

    if ( n >= NBand + 1 ) then
!!$      write( 6, * ) 'Given wavenumber is out of given bands: ', &
!!$        & WaveNum, a_WNBnds(NBand)
!!$      stop
      n = -999
    end if

    BandNum = n

  end function GetBandNum

  !----------------------------------------------------------------------------

  subroutine SetupBin( &
    & NWaveNum, w_WaveNum,  &
    & NBand, b_BandNBin, ab_BandWNBnds, &
    & NGaussReg, a_GaussRegBnds, cb_BandGaussRegNBin, &
    & NBin, &
    & c_BinNum, ab_BandBinIndexBnds, c_BinPointWN, c_BinWeight, &
    & ac_BinWNIndexBnds &
    & )

    use gauss_quad, only : GauLeg

    integer, intent(in ) :: NWaveNum
    real(8), intent(in ) :: w_WaveNum(NWaveNum)
    integer, intent(in ) :: NBand
    integer, intent(in ) :: b_BandNBin(NBand)
                           ! Number of bins in bands
    real(8), intent(in ) :: ab_BandWNBnds(2,NBand)
                           ! Wavenumber bounds of bands
    integer, intent(in ) :: NGaussReg
    real(8), intent(in ) :: a_GaussRegBnds(0:NGaussReg)
    integer, intent(in ) :: cb_BandGaussRegNBin(NGaussReg,NBand)
    integer, intent(in ) :: NBin
                           ! Number of bins
    integer, intent(out) :: c_BinNum    (NBin)
                           ! Bin number
    integer, intent(out) :: ab_BandBinIndexBnds(2,NBand)
                           ! Index bounds of bins in bands
    real(8), intent(out) :: c_BinPointWN(NBin)
                           ! Bin point wavenumber
    real(8), intent(out) :: c_BinWeight (NBin)
                           ! Bin weight
    integer, intent(out) :: ac_BinWNIndexBnds(2,NBin)
                           ! Wavenumber index bounds of bins


    ! Local variables
    !
    real(8) :: c_BinWNBnds      (0:NBin)
                           ! Wavenumber bounds of bins

    real(8) :: WNs
    real(8) :: WNe
    integer :: NSubBin

    integer :: j
    integer :: l
    integer :: m
    integer :: n

    integer :: ls
    integer :: le


    do j = 1, NBin
      c_BinNum(j) = j
    end do
    le = 0
    do n = 1, NBand
      ls = le + 1
      le = ls + b_BandNBin(n) - 1
      ab_BandBinIndexBnds(1,n) = ls
      ab_BandBinIndexBnds(2,n) = le
    end do
    !
    c_BinWeight = 0.0d0
    do n = 1, NBand

      ! 1 region
      !
!!$      ls = ab_BandBinIndexBnds(1,n)
!!$      le = ab_BandBinIndexBnds(2,n)
!!$      WNs = ab_BandWNBnds(1,n)
!!$      WNe = ab_BandWNBnds(2,n)
!!$      NSubBin = b_BandNBin(n)
!!$      call GauLeg( WNs, WNe, &
!!$        & NSubBin, c_BinPointWN(ls:le), c_BinWeight(ls:le) )

      ! 2 regions
      !   1st: from 0 to 0.98
      !   2nd: from 0.98 to 1
      !
!!$      ls = ab_BandBinIndexBnds(1,n)
!!$      le = ls + ( ab_BandBinIndexBnds(2,n) - ab_BandBinIndexBnds(1,n) ) / 2
!!$      WNs = ab_BandWNBnds(1,n)
!!$      WNe = WNs + ( ab_BandWNBnds(2,n) - ab_BandWNBnds(1,n) ) * 0.98d0
!!$      NSubBin = b_BandNBin(n) / 2
!!$      call GauLeg( WNs, WNe, &
!!$        & NSubBin, c_BinPointWN(ls:le), c_BinWeight(ls:le) )
!!$      ls = le + 1
!!$      le = ab_BandBinIndexBnds(2,n)
!!$      WNs = WNe
!!$      WNe = ab_BandWNBnds(2,n)
!!$      NSubBin = b_BandNBin(n) - NSubBin
!!$      call GauLeg( WNs, WNe, &
!!$        & NSubBin, c_BinPointWN(ls:le), c_BinWeight(ls:le) )


      ! N regions
      !
!!$      if ( mod( b_BandNBin(n), NGaussReg ) /= 0 ) then
!!$        write( 6, * ) "Error: NGaussReg is inappropriate: ", n, b_BandNBin(n), NGaussReg
!!$        stop
!!$      end if
!!$      NSubBin = b_BandNBin(n) / NGaussReg
      le = ab_BandBinIndexBnds(1,n) - 1
      WNe = ab_BandWNBnds(1,n)
      do m = 1, NGaussReg
        NSubBin = cb_BandGaussRegNBin(m,n)
        ls = le + 1
        le = ls + NSubBin - 1
        !
        WNs = WNe
        WNe = WNs &
          & + ( ab_BandWNBnds(2,n) - ab_BandWNBnds(1,n) ) &
          &   * ( a_GaussRegBnds(m) - a_GaussRegBnds(m-1) )
        !
        call GauLeg( WNs, WNe, &
          & NSubBin, c_BinPointWN(ls:le), c_BinWeight(ls:le) )
      end do

    end do

    ! Actually, bin is not bounded by wavenumber, but is bounded by 
    ! cumulative probability.
    ! But, here, wavenumber bound is considered for clarity.
    !
    c_BinWNBnds(0) = ab_BandWNBnds(1,1)
    do j = 1, NBin
      c_BinWNBnds(j) = c_BinWNBnds(j-1) + c_BinWeight(j)
    end do

    ! Look for bin bound index
    search_bin_bound_index_1 : do l = 1, NWaveNum
      if ( w_WaveNum(l) >= c_BinWNBnds(0) ) exit search_bin_bound_index_1
    end do search_bin_bound_index_1
    ac_BinWNIndexBnds(1,1) = l
    !
    do j = 1, NBin
      l = NWaveNum
      if ( w_WaveNum(l) <= c_BinWNBnds(j) ) then
        ac_BinWNIndexBnds(2,j) = l
      else
        search_bin_bound_index : do l = ac_BinWNIndexBnds(1,j), NWaveNum
          if ( w_WaveNum(l) > c_BinWNBnds(j) ) exit search_bin_bound_index
        end do search_bin_bound_index
        ac_BinWNIndexBnds(2,j) = l-1
      end if
      if ( j < NBin ) then
        ac_BinWNIndexBnds(1,j+1) = ac_BinWNIndexBnds(2,j) + 1
      end if
    end do


    ! Check
    do n = 1, NBand
      ls = ab_BandBinIndexBnds(1,n)
      le = ab_BandBinIndexBnds(2,n)
      do j = ls, le
        if ( ac_BinWNIndexBnds(2,j) - ac_BinWNIndexBnds(1,j) + 1 <= 0 ) then
          write( 6, * ) "WARNING: Number of values in Band #", n, &
            & " Bin #", j-ls+1, " (total Bin #", j, ") is zero."
        end if
      end do
    end do

  end subroutine SetupBin

  !----------------------------------------------------------------------------

end module sortac_module
